#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2012-2016 Étienne Loks  <etienne.loks_AT_peacefrogsDOTnet>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# See the file COPYING for details.

import datetime
from itertools import groupby

from django.conf import settings
from django.contrib.gis.db import models
from django.core.cache import cache
from django.core.urlresolvers import reverse
from django.db.models import Q, Count, Sum, Max, Avg
from django.db.models.signals import post_save, m2m_changed, post_delete
from django.forms import ValidationError
from django.utils.translation import ugettext_lazy as _, ugettext

from ishtar_common.utils import cached_label_changed, \
    force_cached_label_changed, get_cache, mode

from ishtar_common.models import GeneralType, BaseHistorizedItem, \
    HistoricalRecords, LightHistorizedItem, OwnPerms, Department, Source,\
    SourceType, Person, Organization, Town, Dashboard, IshtarUser, ValueGetter,\
    DocumentTemplate, ShortMenuItem, DashboardFormItem, GeneralRelationType,\
    GeneralRecordRelations, post_delete_record_relation, OperationType, \
    get_external_id, ImageModel, post_save_cache


class RemainType(GeneralType):
    class Meta:
        verbose_name = _(u"Remain type")
        verbose_name_plural = _(u"Remain types")
        ordering = ('label',)
post_save.connect(post_save_cache, sender=RemainType)
post_delete.connect(post_save_cache, sender=RemainType)


class Period(GeneralType):
    order = models.IntegerField(_(u"Order"))
    start_date = models.IntegerField(_(u"Start date"))
    end_date = models.IntegerField(_(u"End date"))
    parent = models.ForeignKey("Period", verbose_name=_(u"Parent period"),
                               blank=True, null=True)

    class Meta:
        verbose_name = _(u"Type Period")
        verbose_name_plural = _(u"Types Period")
        ordering = ('order',)

    def __unicode__(self):
        return self.label
post_save.connect(post_save_cache, sender=Period)
post_delete.connect(post_save_cache, sender=Period)


class ReportState(GeneralType):
    order = models.IntegerField(_(u"Order"))

    class Meta:
        verbose_name = _(u"Type of report state")
        verbose_name_plural = _(u"Types of report state")
        ordering = ('order',)

    def __unicode__(self):
        return self.label
post_save.connect(post_save_cache, sender=ReportState)
post_delete.connect(post_save_cache, sender=ReportState)


class ArchaeologicalSite(BaseHistorizedItem):
    reference = models.CharField(_(u"Reference"), max_length=20, unique=True)
    name = models.CharField(_(u"Name"), max_length=200,
                            null=True, blank=True)
    periods = models.ManyToManyField(Period, verbose_name=_(u"Periods"),
                                     blank=True, null=True)
    remains = models.ManyToManyField("RemainType", verbose_name=_(u'Remains'),
                                     blank=True, null=True)

    class Meta:
        verbose_name = _(u"Archaeological site")
        verbose_name_plural = _(u"Archaeological sites")
        permissions = (
            ("view_archaeologicalsite",
             ugettext(u"Can view all Archaeological sites")),
            ("view_own_archaeologicalsite",
             ugettext(u"Can view own Archaeological site")),
            ("add_own_archaeologicalsite",
             ugettext(u"Can add own Archaeological site")),
            ("change_own_archaeologicalsite",
             ugettext(u"Can change own Archaeological site")),
            ("delete_own_archaeologicalsite",
             ugettext(u"Can delete own Archaeological site")),
        )

    def __unicode__(self):
        name = self.reference
        if self.name:
            name += u" %s %s" % (settings.JOINT, self.name)
        if self.remains.count():
            name += u" {} {}".format(
                settings.JOINT,
                u", ".join([unicode(remain) for remain in self.remains.all()]))
        if self.periods.count():
            name += u" [{}]".format(
                u", ".join([unicode(period) for period in self.periods.all()]))
        return name


def get_values_town_related(item, prefix, values):
    values[prefix + 'parcellist'] = item.render_parcels()
    values[prefix + 'towns'] = ''
    values[prefix + 'departments'] = ''
    values[prefix + 'departments_number'] = ''
    values[prefix + 'towns_count'] = unicode(item.towns.count())
    if item.towns.count():
        values[prefix + 'towns'] = u", ".join([
            town.name for town in item.towns.all().order_by('name')])
        if settings.COUNTRY == 'fr':
            dpts_num = set(
                [town.numero_insee[:2] for town in item.towns.all()])
            values[prefix + 'departments_number'] = u", ".join(
                list(sorted(dpts_num)))
            values[prefix + 'departments'] = u", ".join(
                [Department.objects.get(number=dpt).label
                 for dpt in sorted(dpts_num) if Department.objects.filter(
                    number=dpt).count()])
    return values

QUALITY = (('ND', _(u"Not documented")),
           ('A', _(u"Arbitrary")),
           ('R', _(u"Reliable")),)


class ClosedItem(object):
    def closing(self):
        if self.is_active():
            return
        in_history = False
        date = self.end_date
        # last action is closing?
        for idx, item in enumerate(
                self.history.order_by('-history_date').all()):
            if not idx:
                # last action
                continue
            if not item.end_date or item.end_date != self.end_date:
                break
            in_history = True
        user = None
        if in_history:
            if item.history_modifier_id:
                q = IshtarUser.objects.filter(pk=item.history_modifier_id)
                if q.count():
                    user = q.all()[0]
        elif self.history_modifier_id:
            q = IshtarUser.objects.filter(pk=self.history_modifier_id)
            if q.count():
                user = q.all()[0]
        return {'date': date, 'user': user}


class Operation(ClosedItem, BaseHistorizedItem, ImageModel, OwnPerms,
                ValueGetter, ShortMenuItem, DashboardFormItem):
    QUALITY_DICT = dict(QUALITY)
    SHOW_URL = 'show-operation'
    TABLE_COLS = ['year', 'towns', 'common_name', 'operation_type',
                  'start_date', 'excavation_end_date', 'remains']
    IMAGE_PREFIX = 'operations/'
    SLUG = 'operation'

    # search parameters
    BOOL_FIELDS = ['end_date__isnull', 'virtual_operation',
                   'documentation_received', 'finds_received']
    DATED_FIELDS = [
        'start_date__lte', 'start_date__gte', 'excavation_end_date__lte',
        'excavation_end_date__gte', 'documentation_deadline__lte',
        'documentation_deadline__gte', 'finds_deadline__lte',
        'finds_deadline__gte']
    RELATIVE_SESSION_NAMES = [('file', 'associated_file__pk')]
    EXTRA_REQUEST_KEYS = {
        'common_name': 'common_name__icontains',
        'cached_label': 'cached_label__icontains',
        'comment': 'comment__icontains',
        'scientific_documentation_comment':
        'scientific_documentation_comment__icontains',
        'abstract': 'abstract__icontains',
        'end_date': 'end_date__isnull',
        'start_before': 'start_date__lte',
        'start_after': 'start_date__gte',
        'end_before': 'excavation_end_date__lte',
        'end_after': 'excavation_end_date__gte',
        'towns__numero_insee__startswith':
        'towns__numero_insee__startswith',
        'parcel_0': ('parcels__section',
                     'associated_file__parcels__section'),
        'parcel_1': (
            'parcels__parcel_number',
            'associated_file__parcels__parcel_number'),
        'parcel_2': (
            'parcels__public_domain',
            'associated_file__parcels__public_domain'),
        'history_creator':
        'history_creator__ishtaruser__person__pk',
        'history_modifier':
        'history_modifier__ishtaruser__person__pk',
        'archaeological_sites':
        'archaeological_sites__pk',
        'documentation_deadline_before': 'documentation_deadline__lte',
        'documentation_deadline_after': 'documentation_deadline__gte',
        'finds_deadline_before': 'finds_deadline__lte',
        'finds_deadline_after': 'finds_deadline__gte',
        'related_treatment':
        'context_record__base_finds__find__upstream_treatment__id'
    }

    COL_LABELS = {
        'full_code_patriarche': u"Code patriarche",
        'associated_file_short_label': _(u"Associated file (label)"),
        'operator__name': _(u"Operator name"),
        'scientist__raw_name': _(u"Scientist (full name)"),
        'associated_file__external_id': _(u"Associated file (external ID)"),
        'scientist__title': _(u"Scientist (title)"),
        'scientist__surname': _(u"Scientist (surname)"),
        'scientist__name': _(u"Scientist (name)"),
        'scientist__attached_to__name': _(u"Scientist - Organization (name)"),
        'in_charge__title': _(u"In charge (title)"),
        'in_charge__surname': _(u"In charge (surname)"),
        'in_charge__name': _(u"In charge (name)"),
        'in_charge__attached_to__name': _(u"In charge - Organization (name)"),
        'cira_rapporteur__surname': u"Rapporteur CIRA (prénom)",
        'cira_rapporteur__name': u"Rapporteur CIRA (nom)",
        'cira_rapporteur__attached_to__name': u"Rapporteur CIRA - "
                                              u"Organisation (nom)",
        'archaeological_sites__reference': _(u"Archaeological sites ("
                                             u"reference)"),
    }

    # fields definition
    creation_date = models.DateField(_(u"Creation date"),
                                     default=datetime.date.today)
    end_date = models.DateField(_(u"Closing date"), null=True, blank=True)
    start_date = models.DateField(_(u"Start date"), null=True, blank=True)
    excavation_end_date = models.DateField(
        _(u"Excavation end date"), null=True, blank=True)
    report_delivery_date = models.DateField(_(u"Report delivery date"),
                                            null=True, blank=True)
    scientist = models.ForeignKey(
        Person, blank=True, null=True, verbose_name=_(u"In charge scientist"),
        on_delete=models.SET_NULL,
        related_name='operation_scientist_responsability')
    operator = models.ForeignKey(
        Organization, blank=True, null=True, related_name='operator',
        verbose_name=_(u"Operator"), on_delete=models.SET_NULL)
    in_charge = models.ForeignKey(Person, blank=True, null=True,
                                  verbose_name=_(u"In charge"),
                                  on_delete=models.SET_NULL,
                                  related_name='operation_responsability')
    collaborators = models.ManyToManyField(
        Person, blank=True, null=True, verbose_name=_(u"Collaborators"),
        related_name='operation_collaborator'
    )
    year = models.IntegerField(_(u"Year"), null=True, blank=True)
    operation_code = models.IntegerField(_(u"Numeric reference"), null=True,
                                         blank=True)
    associated_file = models.ForeignKey(
        'archaeological_files.File',
        related_name='operations', verbose_name=_(u"File"),
        blank=True, null=True)
    operation_type = models.ForeignKey(OperationType, related_name='+',
                                       verbose_name=_(u"Operation type"))
    surface = models.IntegerField(_(u"Surface (m2)"), blank=True, null=True)
    remains = models.ManyToManyField("RemainType", verbose_name=_(u'Remains'),
                                     null=True, blank=True)
    towns = models.ManyToManyField(Town, verbose_name=_(u"Towns"),
                                   related_name='operations')
    cost = models.IntegerField(_(u"Cost (euros)"),
                               blank=True, null=True)  # preventive
    periods = models.ManyToManyField(Period, verbose_name=_(u"Periods"),
                                     null=True, blank=True)
    # preventive
    scheduled_man_days = models.IntegerField(_(u"Scheduled man-days"),
                                             blank=True, null=True)
    # preventive
    optional_man_days = models.IntegerField(_(u"Optional man-days"),
                                            blank=True, null=True)
    # preventive
    effective_man_days = models.IntegerField(_(u"Effective man-days"),
                                             blank=True, null=True)
    report_processing = models.ForeignKey(
        ReportState, verbose_name=_(u"Report processing"),
        blank=True, null=True)
    old_code = models.CharField(_(u"Old code"), max_length=200, null=True,
                                blank=True)
    if settings.COUNTRY == 'fr':
        code_patriarche = models.IntegerField(u"Code PATRIARCHE", null=True,
                                              blank=True, unique=True)
        TABLE_COLS = ['full_code_patriarche'] + TABLE_COLS
        # preventive
        fnap_financing = models.FloatField(u"Financement FNAP (%)",
                                           blank=True, null=True)
        # preventive
        fnap_cost = models.IntegerField(u"Financement FNAP (€)",
                                        blank=True, null=True)
        # preventive diag
        zoning_prescription = models.NullBooleanField(
            _(u"Prescription on zoning"), blank=True, null=True)
        # preventive diag
        large_area_prescription = models.NullBooleanField(
            _(u"Prescription on large area"), blank=True, null=True)
        geoarchaeological_context_prescription = models.NullBooleanField(
            _(u"Prescription on geoarchaeological context"), blank=True,
            null=True)  # preventive diag
        cira_rapporteur = models.ForeignKey(
            Person, related_name='cira_rapporteur', null=True, blank=True,
            on_delete=models.SET_NULL, verbose_name=u"Rapporteur CIRA")
        negative_result = models.NullBooleanField(
            u"Résultat considéré comme négatif", blank=True, null=True)
        cira_date = models.DateField(u"Date avis CIRA", null=True, blank=True)
        eas_number = models.CharField(u"Numéro de l'EA", max_length=20,
                                      null=True, blank=True)
    operator_reference = models.CharField(
        _(u"Operator reference"), max_length=20, null=True, blank=True)
    common_name = models.TextField(_(u"Generic name"), null=True, blank=True)
    address = models.TextField(_(u"Address / Locality"), null=True, blank=True)
    comment = models.TextField(_(u"General comment"), null=True, blank=True)
    scientific_documentation_comment = models.TextField(
        _(u"Comment about scientific documentation"), null=True, blank=True)
    cached_label = models.CharField(_(u"Cached name"), max_length=500,
                                    null=True, blank=True)
    archaeological_sites = models.ManyToManyField(
        ArchaeologicalSite, verbose_name=_(u"Archaeological sites"),
        null=True, blank=True)
    virtual_operation = models.BooleanField(
        _(u"Virtual operation"),
        default=False, help_text=_(
            u"If checked, it means that this operation have not been "
            u"officialy registered."))
    record_quality = models.CharField(
        _(u"Record quality"), max_length=2, null=True, blank=True,
        choices=QUALITY)
    abstract = models.TextField(_(u"Abstract"), null=True, blank=True)
    documentation_deadline = models.DateField(
        _(u"Deadline for submission of the documentation"), blank=True,
        null=True)
    documentation_received = models.NullBooleanField(
        _(u"Documentation received"), blank=True, null=True)
    finds_deadline = models.DateField(
        _(u"Deadline for submission of the finds"), blank=True, null=True)
    finds_received = models.NullBooleanField(
        _(u"Finds received"), blank=True, null=True)

    point = models.PointField(_(u"Point"), blank=True, null=True)
    multi_polygon = models.MultiPolygonField(_(u"Multi polygon"), blank=True,
                                             null=True)
    history = HistoricalRecords()

    class Meta:
        verbose_name = _(u"Operation")
        verbose_name_plural = _(u"Operations")
        permissions = (
            ("view_operation", ugettext(u"Can view all Operations")),
            ("view_own_operation", ugettext(u"Can view own Operation")),
            ("add_own_operation", ugettext(u"Can add own Operation")),
            ("change_own_operation", ugettext(u"Can change own Operation")),
            ("delete_own_operation", ugettext(u"Can delete own Operation")),
            ("close_operation", ugettext(u"Can close Operation")),
        )
        ordering = ('cached_label',)

    @classmethod
    def get_owns(cls, user, menu_filtr=None, limit=None, values=None,
                 get_short_menu_class=None):
        replace_query = None
        if menu_filtr and 'file' in menu_filtr:
            replace_query = Q(associated_file=menu_filtr['file'])

        owns = super(Operation, cls).get_owns(
            user, replace_query=replace_query,
            limit=limit, values=values,
            get_short_menu_class=get_short_menu_class)
        return cls._return_get_owns(owns, values, get_short_menu_class)

    def __unicode__(self):
        if self.cached_label:
            return self.cached_label
        self.save()
        return self.cached_label

    def get_values(self, prefix=''):
        values = super(Operation, self).get_values(prefix=prefix)
        return get_values_town_related(self, prefix, values)

    @property
    def short_class_name(self):
        return _(u"OPE")

    @property
    def external_id(self):
        return self.code_patriarche

    @property
    def short_label(self):
        if settings.COUNTRY == 'fr':
            return self.reference
        return unicode(self)

    @property
    def name(self):
        return self.common_name

    @property
    def show_url(self):
        return reverse('show-operation', args=[self.pk, ''])

    def towns_codes(self):
        return [u"{} ({})".format(town.name, town.numero_insee) for town in
                self.towns.all()]

    def has_finds(self):
        from archaeological_finds.models import BaseFind
        return BaseFind.objects.filter(context_record__operation=self).count()

    def finds(self):
        from archaeological_finds.models import BaseFind
        return BaseFind.objects.filter(context_record__operation=self)

    def get_reference(self, full=False):
        ref = ""
        if settings.COUNTRY == 'fr' and self.code_patriarche:
            ref = settings.ISHTAR_OPE_PREFIX + unicode(self.code_patriarche)
            if not full:
                return ref
        if self.year and self.operation_code:
            if ref:
                ref += u" - "
            ref += settings.ISHTAR_DEF_OPE_PREFIX
            ref += u"-".join((unicode(self.year),
                              unicode(self.operation_code)))
        return ref or "00"

    @property
    def reference(self):
        return self.get_reference()

    @property
    def report_delivery_delay(self):
        return None
        # q = self.source.filter(source_type__txt_idx__endswith='_report')
        # if not self.report_delivery_date or not q.count():
        #     return None

    def _generate_cached_label(self):
        items = [self.get_town_label(), self.get_reference(full=True)]
        if self.common_name:
            items.append(self.common_name)
        cached_label = settings.JOINT.join(items)
        return cached_label

    def _get_associated_cached_labels(self):
        return list(self.context_record.all())

    def _cached_labels_bulk_update(self):
        if settings.TESTING and settings.USE_SPATIALITE_FOR_TESTS:
            return
        self.context_record.model.cached_label_bulk_update(operation_id=self.pk)
        return True

    def get_town_label(self):
        lbl = unicode(_('Intercommunal'))
        if self.towns.count() == 1:
            lbl = self.towns.all()[0].name
        return lbl

    def get_department(self):
        if not self.towns.count():
            return '00'
        return self.towns.all()[0].numero_insee[:2]

    def grouped_parcels(self):
        return Parcel.grouped_parcels(list(self.parcels.all()))

    def render_parcels(self):
        return Parcel.render_parcels(list(self.parcels.all()))

    def context_record_relations_q(self):
        from archaeological_context_records.models \
            import RecordRelations as CRRL
        return CRRL.objects.filter(left_record__operation=self)

    def context_record_docs_q(self):
        from archaeological_context_records.models import ContextRecordSource
        return ContextRecordSource.objects.filter(
            context_record__operation=self)

    def find_docs_q(self):
        from archaeological_finds.models import FindSource
        return FindSource.objects.filter(
            find__base_finds__context_record__operation=self)

    def containers_q(self):
        from archaeological_warehouse.models import Container
        return Container.objects.filter(
            finds__base_finds__context_record__operation=self
        )

    associated_file_short_label_lbl = _(u"Archaeological file")
    full_code_patriarche_lbl = _(u"Code patriarche")

    @property
    def associated_file_short_label(self):
        if not self.associated_file:
            return ""
        return self.associated_file.short_label

    @classmethod
    def get_available_operation_code(cls, year=None):
        max_val = cls.objects.filter(year=year).aggregate(
            Max('operation_code'))["operation_code__max"]
        return (max_val + 1) if max_val else 1

    year_index_lbl = _(u"Operation code")

    @property
    def year_index(self):
        if not self.operation_code:
            return ""
        lbl = unicode(self.operation_code)
        year = self.year or 0
        lbl = settings.ISHTAR_DEF_OPE_PREFIX \
              + u"%d-%s%s" % (year, (3 - len(lbl)) * "0", lbl)
        return lbl

    @property
    def full_code_patriarche(self):
        if not self.code_patriarche:
            return ''
        return settings.ISHTAR_OPE_PREFIX + unicode(self.code_patriarche)

    def clean(self):
        if not self.operation_code:
            return
        objs = self.__class__.objects.filter(
            year=self.year, operation_code=self.operation_code)
        if self.pk:
            objs = objs.exclude(pk=self.pk)
        if objs.count():
            raise ValidationError(_(u"This operation code already exists for "
                                    u"this year"))

    @property
    def surface_ha(self):
        if self.surface:
            return self.surface / 10000.0

    @property
    def cost_by_m2(self):
        if not self.surface or not self.cost:
            return
        return round(float(self.cost) / self.surface, 2)

    @classmethod
    def get_query_owns(cls, user):
        return (
            Q(in_charge=user.ishtaruser.person) |
            Q(scientist=user.ishtaruser.person) |
            Q(collaborators__pk=user.ishtaruser.person.pk) |
            Q(history_creator=user)) & Q(end_date__isnull=True)

    def is_active(self):
        return not bool(self.end_date)

    @property
    def nb_parcels(self):
        _(u"Number of parcels")
        nb = 0
        if self.associated_file:
            nb = self.associated_file.parcels.count()
        if not nb:
            nb = self.parcels.count()
        return nb

    @property
    def nb_acts(self, update=False):
        _(u"Number of administrative acts")
        return self._get_or_set_stats('_nb_acts', update)

    def _nb_acts(self):
        return self.administrative_act.count()

    @property
    def nb_indexed_acts(self, update=False):
        _(u"Number of indexed administrative acts")
        return self._get_or_set_stats('_nb_indexed_acts', update)

    def _nb_indexed_acts(self):
        return self.administrative_act.filter(act_type__indexed=True).count()

    @property
    def nb_context_records(self, update=False):
        _(u"Number of context records")
        return self._get_or_set_stats('_nb_context_records', update)

    def _nb_context_records(self):
        return self.context_record.count()

    @property
    def nb_context_records_by_type(self, update=False):
        return self._get_or_set_stats('_nb_context_records_by_type', update)

    def _nb_context_records_by_type(self):
        nbs = []
        q = self.context_record.values(
            'unit', 'unit__label').distinct().order_by('label')
        for res in q.all():
            nbs.append((unicode(res['unit__label']),
                        self.context_record.filter(unit=res['unit']).count()))
        return list(set(nbs))

    @property
    def nb_context_records_by_periods(self, update=False):
        return self._get_or_set_stats('_nb_context_records_by_periods', update)

    def _nb_context_records_by_periods(self):
        nbs = []
        q = self.context_record.values(
            'datings__period', 'datings__period__label').distinct().order_by(
                'datings__period__order')
        for res in q.all():
            nbs.append((unicode(res['datings__period__label']),
                        self.context_record.filter(
                            datings__period=res['datings__period']).count()))
        return nbs

    @property
    def nb_finds(self, update=False):
        _(u"Number of finds")
        return self._get_or_set_stats('_nb_finds', update)

    def _nb_finds(self):
        from archaeological_finds.models import Find
        q = Find.objects.filter(
            base_finds__context_record__operation=self,
            upstream_treatment_id__isnull=True).distinct()
        return q.count()

    @property
    def nb_finds_by_material_type(self, update=False):
        return self._get_or_set_stats('_nb_finds_by_material_type', update)

    def _nb_finds_by_material_type(self):
        from archaeological_finds.models import Find
        nbs = []
        q = Find.objects.filter(
            upstream_treatment_id__isnull=True,
            base_finds__context_record__operation=self).distinct().values(
            'material_types__pk', 'material_types__label').distinct().order_by(
            'material_types__label')
        for res in q.all():
            nbs.append(
                (unicode(res['material_types__label']),
                 Find.objects.filter(
                    base_finds__context_record__operation=self,
                    upstream_treatment_id__isnull=True,
                    material_types__pk=res['material_types__pk']).count()))
        return nbs

    @property
    def nb_finds_by_types(self, update=False):
        return self._get_or_set_stats('_nb_finds_by_types', update)

    def _nb_finds_by_types(self):
        from archaeological_finds.models import Find
        nbs = []
        q = Find.objects.filter(
            base_finds__context_record__operation=self).values(
            'object_types', 'object_types__label').distinct().order_by(
                'object_types__label')
        for res in q.all():
            label = unicode(res['object_types__label'])
            if label == 'None':
                label = _(u"No type")
            nbs.append(
                (label,
                 Find.objects.filter(
                     base_finds__context_record__operation=self,
                     upstream_treatment_id__isnull=True,
                     object_types=res['object_types']).count()))
        return nbs

    @property
    def nb_finds_by_periods(self, update=False):
        return self._get_or_set_stats('_nb_finds_by_periods', update)

    def _nb_finds_by_periods(self):
        from archaeological_finds.models import Find
        nbs = []
        q = Find.objects.filter(
            base_finds__context_record__operation=self).values(
            'datings__period', 'datings__period__label').distinct().order_by(
                'datings__period__order')
        for res in q.all():
            nbs.append(
                (unicode(res['datings__period__label']),
                 Find.objects.filter(
                    base_finds__context_record__operation=self,
                    upstream_treatment_id__isnull=True,
                    datings__period=res['datings__period']).count()))
        return nbs

    @property
    def nb_documents(self, update=False):
        _(u"Number of sources")
        return self._get_or_set_stats('_nb_documents', update)

    def _nb_documents(self):
        from archaeological_context_records.models import ContextRecordSource
        from archaeological_finds.models import FindSource
        nbs = self.source.count() + \
            ContextRecordSource.objects.filter(
                context_record__operation=self).count() + \
            FindSource.objects.filter(
                find__base_finds__context_record__operation=self).count()
        return nbs

    @property
    def nb_documents_by_types(self, update=False):
        return self._get_or_set_stats('_nb_documents_by_types', update)

    def _nb_documents_by_types(self):
        from archaeological_context_records.models import ContextRecordSource
        from archaeological_finds.models import FindSource
        docs = {}

        qs = [
            self.source,
            ContextRecordSource.objects.filter(context_record__operation=self),
            FindSource.objects.filter(
                find__upstream_treatment_id__isnull=True,
                find__base_finds__context_record__operation=self)]
        for q in qs:
            for res in q.values('source_type').distinct():
                st = res['source_type']
                if st not in docs:
                    docs[st] = 0
                docs[st] += q.filter(source_type=st).count()
        docs = [(unicode(SourceType.objects.get(pk=k)), docs[k]) for k in docs]
        return sorted(docs, key=lambda x: x[0])

    @property
    def nb_stats_finds_by_ue(self, update=False):
        return self._get_or_set_stats('_nb_stats_finds_by_ue', update)

    def _nb_stats_finds_by_ue(self):
        _(u"Mean")
        res, finds = {}, []
        for cr in self.context_record.all():
            finds.append(cr.base_finds.count())
        if not finds:
            return res
        res['mean'] = float(sum(finds)) / max(len(finds), 1)
        res['min'] = min(finds)
        res['max'] = max(finds)
        res['mode'] = u" ; ".join([str(m) for m in mode(finds)])
        return res

    def save(self, *args, **kwargs):
        # put a default year if start_date is defined
        if self.start_date and not self.year:
            self.year = self.start_date.year
        if self.operation_code is None:
            self.operation_code = self.get_available_operation_code(self.year)
        return super(Operation, self).save(*args, **kwargs)


m2m_changed.connect(force_cached_label_changed, sender=Operation.towns.through)


def operation_post_save(sender, **kwargs):
    if not kwargs['instance']:
        return
    operation = kwargs['instance']
    operation.skip_history_when_saving = True
    if operation.fnap_financing and operation.cost:
        fnap_cost = int(float(operation.cost) / 100 * operation.fnap_financing)
        if not operation.fnap_cost or operation.fnap_cost != fnap_cost:
            operation.fnap_cost = fnap_cost
            operation.save()
    elif operation.fnap_cost and operation.cost:
        fnap_percent = float(operation.fnap_cost) * 100 / operation.cost
        if operation.fnap_financing != fnap_percent:
            operation.fnap_financing = fnap_percent
            operation.save()
    cached_label_changed(sender, **kwargs)
    if operation.associated_file:
        operation.associated_file.update_short_menu_class()
        # manage parcel association
        for parcel in operation.parcels.all():
            parcel.copy_to_file()
post_save.connect(operation_post_save, sender=Operation)


class RelationType(GeneralRelationType):
    inverse_relation = models.ForeignKey(
        'RelationType', verbose_name=_(u"Inverse relation"), blank=True,
        null=True)

    class Meta:
        verbose_name = _(u"Operation relation type")
        verbose_name_plural = _(u"Operation relation types")
        ordering = ('order', 'label')


class RecordRelations(GeneralRecordRelations, models.Model):
    MAIN_ATTR = 'left_record'
    left_record = models.ForeignKey(Operation,
                                    related_name='right_relations')
    right_record = models.ForeignKey(Operation,
                                     related_name='left_relations')
    relation_type = models.ForeignKey(RelationType)

    class Meta:
        verbose_name = _(u"Operation record relation")
        verbose_name_plural = _(u"Operation record relations")
        ordering = ('left_record', 'relation_type')

post_delete.connect(post_delete_record_relation, sender=RecordRelations)


class OperationByDepartment(models.Model):
    """
    Database view for dashboard
    """
    operation = models.ForeignKey(Operation, verbose_name=_(u"Operation"))
    department = models.ForeignKey(Department, verbose_name=_(u"Department"),
                                   blank=True, null=True)

    class Meta:
        managed = False
        db_table = 'operation_department'


class OperationSource(Source):
    SHOW_URL = 'show-operationsource'
    MODIFY_URL = 'operation_source_modify'
    TABLE_COLS = ['operation__code_patriarche', 'operation__year',
                  'operation__operation_code', 'code'] + Source.TABLE_COLS

    # search parameters
    BOOL_FIELDS = ['duplicate']
    EXTRA_REQUEST_KEYS = {
        'title': 'title__icontains',
        'description': 'description__icontains',
        'comment': 'comment__icontains',
        'additional_information': 'additional_information__icontains',
        'person': 'authors__person__pk',
        'operation__towns': 'operation__towns__pk',
        'operation__operation_code': 'operation__operation_code',
        'operation__code_patriarche': 'operation__code_patriarche',
        'operation__operation_type': 'operation__operation_type__pk',
        'operation__year': 'operation__year'}
    COL_LABELS = {
        'operation__year': _(u"Operation year"),
        'operation__operation_code': _(u"Operation code"),
        'code': _(u"Document code")
    }

    # fields
    operation = models.ForeignKey(Operation, verbose_name=_(u"Operation"),
                                  related_name="source")
    index = models.IntegerField(verbose_name=_(u"Index"), blank=True,
                                null=True)

    class Meta:
        verbose_name = _(u"Operation documentation")
        verbose_name_plural = _(u"Operation documentations")
        permissions = (
            ("view_operationsource",
             ugettext(u"Can view all Operation sources")),
            ("view_own_operationsource",
             ugettext(u"Can view own Operation source")),
            ("add_own_operationsource",
             ugettext(u"Can add own Operation source")),
            ("change_own_operationsource",
             ugettext(u"Can change own Operation source")),
            ("delete_own_operationsource",
             ugettext(u"Can delete own Operation source")),
        )

    @property
    def owner(self):
        return self.operation

    @property
    def code(self):
        return u"{}-{:04d}".format(self.operation.code_patriarche or '',
                                   self.index)

    @classmethod
    def get_query_owns(cls, user):
        return (Q(operation__in_charge=user.ishtaruser.person) |
                Q(operation__scientist=user.ishtaruser.person) |
                Q(operation__collaborators__pk=user.ishtaruser.person.pk)) \
               & Q(operation__end_date__isnull=True)


class ActType(GeneralType):
    TYPE = (('F', _(u'Archaeological file')),
            ('O', _(u'Operation')),
            ('TF', _(u'Treatment request')),
            ('T', _(u'Treatment')),
            )
    intented_to = models.CharField(_(u"Intended to"), max_length=2,
                                   choices=TYPE)
    code = models.CharField(_(u"Code"), max_length=10, blank=True, null=True)
    associated_template = models.ManyToManyField(
        DocumentTemplate, blank=True, null=True,
        verbose_name=_(u"Associated template"), related_name='acttypes')
    indexed = models.BooleanField(_(u"Indexed"), default=False)

    class Meta:
        verbose_name = _(u"Act type")
        verbose_name_plural = _(u"Act types")
        ordering = ('label',)
post_save.connect(post_save_cache, sender=ActType)
post_delete.connect(post_save_cache, sender=ActType)


class AdministrativeAct(BaseHistorizedItem, OwnPerms, ValueGetter):
    TABLE_COLS = ['full_ref', 'year', 'index', 'act_type', 'act_object',
                  'signature_date', 'associated_file', 'operation',
                  'towns_label']
    TABLE_COLS_FILE = [
        'full_ref', 'year', 'index', 'act_type',
        'act_object', 'associated_file', 'towns_label',
    ]
    TABLE_COLS_OPE = ['full_ref', 'year', 'index', 'act_type', 'operation',
                      'act_object', 'towns_label']
    if settings.COUNTRY == 'fr':
        TABLE_COLS.append('departments_label')
        TABLE_COLS_FILE.append('departments_label')
        TABLE_COLS_OPE.append('departments_label')

    # search parameters
    DATED_FIELDS = ['signature_date__lte', 'signature_date__gte']
    ASSOCIATED_MODELS = [
        ('File', 'associated_file'),
        (Person, 'associated_file__general_contractor')]
    EXTRA_REQUEST_KEYS = {
        'act_object': 'act_object__icontains',
        'act_type__intented_to': 'act_type__intented_to',
        'associated_file__general_contractor__attached_to':
            'associated_file__general_contractor__attached_to__pk',
        'associated_file__name': 'associated_file__name__icontains',
        'associated_file__operations__code_patriarche':
        'associated_file__operations__code_patriarche',
        'associated_file__permit_reference':
            'associated_file__permit_reference__icontains',
        'associated_file__towns': 'associated_file__towns__pk',
        'associated_file__towns__numero_insee__startswith':
        'associated_file__towns__numero_insee__startswith',
        'indexed': 'index__isnull',
        'history_creator':
        'history_creator__ishtaruser__person__pk',
        'history_modifier':
        'history_modifier__ishtaruser__person__pk',
        'operation__code_patriarche': 'operation__code_patriarche',
        'operation__towns': 'operation__towns__pk',
        'operation__towns__numero_insee__startswith':
        'operation__towns__numero_insee__startswith',
        'parcel_0': ('associated_file__parcels__section',
                     'operation__parcels__section',
                     'operation__associated_file__parcels__section'),
        'parcel_1': (
            'associated_file__parcels__parcel_number'
            'operation__parcels__parcel_number',
            'operation__associated_file__parcels__parcel_number'),
        'parcel_2': (
            'associated_file__parcels__public_domain',
            'operation__parcels__public_domain',
            'operation__associated_file__parcels__public_domain'),
        'signature_date_before': 'signature_date__lte',
        'signature_date_after': 'signature_date__gte',
        'year': 'signature_date__year',
    }
    REVERSED_BOOL_FIELDS = ['index__isnull']
    RELATIVE_SESSION_NAMES = [('operation', 'operation__pk'),
                              ('file', 'associated_file__pk')]
    COL_LABELS = {'full_ref': _(u"Ref.")}

    # fields
    act_type = models.ForeignKey(ActType, verbose_name=_(u"Act type"))
    in_charge = models.ForeignKey(
        Person, blank=True, null=True,
        related_name='adminact_operation_in_charge',
        verbose_name=_(u"Person in charge of the operation"),
        on_delete=models.SET_NULL,)
    index = models.IntegerField(verbose_name=_(u"Index"), blank=True,
                                null=True)
    operator = models.ForeignKey(
        Organization, blank=True, null=True,
        verbose_name=_(u"Archaeological preventive operator"),
        related_name='adminact_operator', on_delete=models.SET_NULL)
    scientist = models.ForeignKey(
        Person, blank=True, null=True,
        related_name='adminact_scientist', on_delete=models.SET_NULL,
        verbose_name=_(u"Scientist in charge"))
    signatory = models.ForeignKey(
        Person, blank=True, null=True, related_name='signatory',
        verbose_name=_(u"Signatory"), on_delete=models.SET_NULL,)
    operation = models.ForeignKey(
        Operation, blank=True, null=True,
        related_name='administrative_act', verbose_name=_(u"Operation"))
    associated_file = models.ForeignKey(
        'archaeological_files.File',
        blank=True, null=True,
        related_name='administrative_act',
        verbose_name=_(u"Archaeological file"))
    treatment_file = models.ForeignKey(
        'archaeological_finds.TreatmentFile',
        blank=True, null=True,
        related_name='administrative_act',
        verbose_name=_(u"Treatment request"))
    treatment = models.ForeignKey(
        'archaeological_finds.Treatment',
        blank=True, null=True,
        related_name='administrative_act',
        verbose_name=_(u"Treatment"))
    signature_date = models.DateField(_(u"Signature date"), blank=True,
                                      null=True)
    year = models.IntegerField(_(u"Year"), blank=True, null=True)
    act_object = models.TextField(_(u"Object"), max_length=300, blank=True,
                                  null=True)
    if settings.COUNTRY == 'fr':
        ref_sra = models.CharField(u"Référence SRA", max_length=15,
                                   blank=True, null=True)
        departments_label = models.TextField(
            _(u"Departments"), blank=True, null=True,
            help_text=_(u"Cached values get from associated departments"))
    towns_label = models.TextField(
        _(u"Towns"), blank=True, null=True,
        help_text=_(u"Cached values get from associated towns"))
    history = HistoricalRecords()
    _prefix = 'adminact_'

    class Meta:
        ordering = ('year', 'signature_date', 'index', 'act_type')
        verbose_name = _(u"Administrative act")
        verbose_name_plural = _(u"Administrative acts")
        permissions = (
            ("view_administrativeact",
             ugettext(u"Can view all Administrative acts")),
            ("view_own_administrativeact",
             ugettext(u"Can view own Administrative act")),
            ("add_own_administrativeact",
             ugettext(u"Can add own Administrative act")),
            ("change_own_administrativeact",
             ugettext(u"Can change own Administrative act")),
            ("delete_own_administrativeact",
             ugettext(u"Can delete own Administrative act")),
        )

    def __unicode__(self):
        return settings.JOINT.join(
            [unicode(item) for item in [
                self.related_item, self.act_object]
             if item])

    full_ref_lbl = _(u"Ref.")

    @property
    def full_ref(self):
        lbl = []
        if self.year:
            lbl.append(unicode(self.year))
        if self.index:
            lbl.append(u"n°%d" % self.index)
        if settings.COUNTRY == 'fr' and self.ref_sra:
            lbl.append(u"[%s]" % self.ref_sra)
        return u" ".join(lbl)

    @property
    def associated_filename(self):
        return self.get_filename()

    @property
    def towns(self):
        if self.associated_file:
            return self.associated_file.towns.all()
        elif self.operation:
            return self.operation.towns.all()
        return []

    @property
    def departments(self):
        if settings.COUNTRY != 'fr':
            return ''
        q = None
        if self.associated_file:
            q = self.associated_file.towns.all()
        elif self.operation:
            q = self.operation.towns.all()
        if not q:
            return ''
        dpts = []
        for town in q:
            dpt = town.numero_insee[:2]
            if dpt not in dpts:
                dpts.append(dpt)
        return ', '.join(list(sorted(dpts)))

    @property
    def related_item(self):
        if self.operation:
            return self.operation
        if self.associated_file:
            return self.associated_file
        if self.treatment:
            return self.treatment
        if self.treatment_file:
            return self.treatment_file

    def get_filename(self):
        filename = self.related_item.associated_filename
        filename = u"-".join(filename.split('-')[:-1])  # remove date
        if self.act_type.code:
            filename += u"-" + self.act_type.code
        if self.signature_date and self.index:
            filename += u"-%d-%d" % (self.signature_date.year,
                                     self.index)
        if self.signature_date:
            filename += u"-" + self.signature_date.strftime('%Y%m%d')
        return filename

    def publish(self, template_pk=None):
        if not self.act_type.associated_template.count():
            return
        if not template_pk:
            template = self.act_type.associated_template.all()[0]
        else:
            q = self.act_type.associated_template.filter(pk=template_pk)
            if not q.count():
                return
            template = q.all()[0]
        return template.publish(self)

    def _get_index(self):
        if not self.index:
            c_index = 1
            q = AdministrativeAct.objects.filter(
                act_type__indexed=True, signature_date__year=self.year,
                index__isnull=False).order_by("-index")
            if q.count():
                c_index = q.all()[0].index + 1
            self.index = c_index
        conflict = AdministrativeAct.objects.filter(
            act_type__indexed=True, signature_date__year=self.year,
            index=self.index)
        if self.pk:
            conflict = conflict.exclude(pk=self.pk)
        if conflict.count():
            if self.pk:
                raise ValidationError(_(u"This index already exists for "
                                        u"this year"))
            else:
                self._get_index()

    def clean(self, *args, **kwargs):
        if not self.signature_date:
            return super(AdministrativeAct, self).clean(*args, **kwargs)
        self.year = self.signature_date.year
        if not self.act_type.indexed:
            return super(AdministrativeAct, self).clean(*args, **kwargs)
        self._get_index()
        super(AdministrativeAct, self).clean(*args, **kwargs)

    def save(self, *args, **kwargs):
        if settings.COUNTRY == 'fr':
            self.departments_label = self.departments
        self.towns_label = u", ".join(
            list(sorted([unicode(town) for town in self.towns])))

        force = False
        if 'force' in kwargs:
            force = kwargs.pop('force')

        if not self.signature_date:
            return super(AdministrativeAct, self).save(*args, **kwargs)
        self.year = self.signature_date.year

        if not self.act_type.indexed:
            return super(AdministrativeAct, self).save(*args, **kwargs)

        if not force:
            self._get_index()
        else:
            try:
                self._get_index()
            except:
                pass

        super(AdministrativeAct, self).save(*args, **kwargs)
        if hasattr(self, 'associated_file') and self.associated_file:
            self.associated_file.update_has_admin_act()
            self.associated_file.update_short_menu_class()


def strip_zero(value):
    for idx, nb in enumerate(value):
        if nb != '0':
            return value[idx:]
    return value


class Parcel(LightHistorizedItem):
    associated_file = models.ForeignKey(
        'archaeological_files.File',
        related_name='parcels', verbose_name=_(u"File"),
        blank=True, null=True, on_delete=models.CASCADE)
    operation = models.ForeignKey(
        Operation, related_name='parcels', blank=True, null=True,
        verbose_name=_(u"Operation"), on_delete=models.CASCADE)
    year = models.IntegerField(_(u"Year"), blank=True, null=True)
    town = models.ForeignKey(Town, related_name='parcels',
                             verbose_name=_(u"Town"))
    section = models.CharField(_(u"Section"), max_length=4,
                               null=True, blank=True)
    parcel_number = models.CharField(_(u"Parcel number"), max_length=6,
                                     null=True, blank=True)
    public_domain = models.BooleanField(_(u"Public domain"), default=False)
    external_id = models.CharField(_(u"External ID"), max_length=100,
                                   null=True, blank=True)
    auto_external_id = models.BooleanField(
        _(u"External ID is set automatically"), default=False)
    address = models.TextField(_(u"Address - Locality"), null=True, blank=True)

    class Meta:
        verbose_name = _(u"Parcel")
        verbose_name_plural = _(u"Parcels")
        ordering = ('year', 'section', 'parcel_number')

    @property
    def short_label(self):
        items = [unicode(item) for item in [self.section, self.parcel_number]
                 if item]
        if self.public_domain:
            items.append(unicode(_(u"Public domain")))
        return settings.JOINT.join(items)

    def __unicode__(self):
        return self.short_label

    """
    def merge(self, parcel):
        # cannot automatically merge
        if self.address and parcel.address and self.address != parcel.address:
            return
        if self.external_id and parcel.external_id and \
           self.external_id != parcel.external_id:
            return
        if self.year and parcel.year and \
           self.year != parcel.year:
            return
        self.address = self.address or parcel.address
        self.external_id = self.external_id or parcel.external_id
        self.year = self.year or parcel.year
        self.save()
        for owner in parcel.owners.all():
            owner.parcel = self
            owner.save()
        if hasattr(parcel, 'context_record'):
            for cr in parcel.context_record.all():
                cr.parcel = self
                cr.save()
        parcel.delete()
    """

    @classmethod
    def grouped_parcels(cls, parcels):
        sortkeyfn = lambda s: (getattr(s, 'town_id'),
                               getattr(s, 'section'), getattr(s, 'year'))
        parcels = sorted(parcels, key=sortkeyfn)
        grouped = []
        for keys, parcel_grp in groupby(parcels, key=sortkeyfn):
            for idx, parcel in enumerate(parcel_grp):
                if not idx:
                    grouped.append(parcel)
                    grouped[-1].parcel_numbers = []
                nb = ""
                if parcel.parcel_number:
                    nb = u"0" * (12 - len(parcel.parcel_number)) + \
                        parcel.parcel_number
                if parcel.public_domain:
                    if nb:
                        nb += " "
                    nb += unicode(_(u"Public domain"))
                grouped[-1].parcel_numbers.append(nb)
            grouped[-1].parcel_numbers.sort()
            grouped[-1].parcel_numbers = [strip_zero(n)
                                          for n in grouped[-1].parcel_numbers]
        return grouped

    @classmethod
    def render_parcels(cls, parcels):
        parcels = cls.grouped_parcels(parcels)
        res = ''
        c_town, c_section = '', ''
        for idx, parcels in enumerate(parcels):
            if c_town != unicode(parcels.town):
                c_town = unicode(parcels.town)
                c_section = ''
                if idx:
                    res += " ; "
                res += unicode(parcels.town) + u' : '
            if c_section:
                res += u" / "
            c_section = parcels.section
            res += parcels.section + u' '
            res += u", ".join(parcels.parcel_numbers)
            if parcels.year:
                res += " (%s)" % unicode(parcels.year)
        return res

    def long_label(self):
        items = [unicode(self.operation) or
                 unicode(self.associated_file) or ""]
        items += [unicode(item) for item in [self.section, self.parcel_number]
                  if item]
        return settings.JOINT.join(items)

    def copy_to_file(self):
        """
        Copy from operation to file when associating file to operation
        """
        if not self.operation or not self.operation.associated_file:
            # not concerned
            return
        keys = {'town': self.town, 'section': self.section,
                'parcel_number': self.parcel_number}
        if self.operation.associated_file.parcels.filter(**keys).count():
            # everything is OK
            return
        keys['address'] = self.address
        keys['year'] = self.year
        keys['associated_file'] = self.operation.associated_file
        new_p = Parcel.objects.create(**keys)
        # also copy owning
        for owning in self.owners.all():
            ParcelOwner.objects.create(
                owner=owning.owner, parcel=new_p,
                start_date=owning.start_date, end_date=owning.end_date)

    def copy_to_operation(self):
        """
        Parcel cannot have operation and associated_file but on
        new parcel association a copy have to be done before cleaning
        """
        if not (self.operation and self.associated_file):
            # everything is OK
            return
        keys = {'town': self.town, 'section': self.section,
                'parcel_number': self.parcel_number,
                'operation': self.operation,
                'associated_file': None,
                'defaults': {'address': self.address, 'year': self.year}
                }
        new_p, created = Parcel.objects.get_or_create(**keys)
        # copy owning only if created
        if created:
            for owning in self.owners.all():
                ParcelOwner.objects.create(
                    owner=owning.owner, parcel=new_p,
                    start_date=owning.start_date, end_date=owning.end_date)
        self.operation = None
        self.save()


def parcel_post_save(sender, **kwargs):
    if not kwargs['instance']:
        return
    parcel = kwargs['instance']
    created = kwargs.get('created', None)

    updated = False
    # remove when the parcel is linked to nothing
    if not getattr(parcel, '_updated_id', None) and not created \
            and not parcel.operation and not parcel.associated_file:
        if parcel.context_record.count():
            # trying to restore a lost parcel
            parcel.operation = parcel.context_record.all()[0].operation
            updated = True
        else:
            parcel.delete()
            return

    if not parcel.external_id or parcel.auto_external_id:
        external_id = get_external_id('parcel_external_id', parcel)
        if external_id != parcel.external_id:
            updated = True
            parcel._updated_id = True
            parcel.auto_external_id = True
            parcel.external_id = external_id
    if updated:
        parcel.save()
        return

    if parcel.context_record.count():
        if settings.TESTING and settings.USE_SPATIALITE_FOR_TESTS:
            for cr in parcel.context_record.all():
                cr.skip_history_when_saving = True
                cr.save()
        else:
            parcel.context_record.model.cached_label_bulk_update(
                parcel_id=parcel.id)

    if parcel.operation and parcel.operation.pk and \
       parcel.town not in list(parcel.operation.towns.all()):
        parcel.operation.towns.add(parcel.town)
    if parcel.associated_file and \
       parcel.associated_file.pk and \
       parcel.town not in list(parcel.associated_file.towns.all()):
        parcel.associated_file.towns.add(parcel.town)
    if parcel.operation and parcel.associated_file:
        # parcels are copied between files and operations
        parcel.copy_to_operation()

post_save.connect(parcel_post_save, sender=Parcel)


class ParcelOwner(LightHistorizedItem):
    owner = models.ForeignKey(Person, verbose_name=_(u"Owner"),
                              related_name="parcel_owner")
    parcel = models.ForeignKey(Parcel, verbose_name=_(u"Parcel"),
                               related_name='owners')
    start_date = models.DateField(_(u"Start date"))
    end_date = models.DateField(_(u"End date"))

    class Meta:
        verbose_name = _(u"Parcel owner")
        verbose_name_plural = _(u"Parcel owners")

    def __unicode__(self):
        return self.owner + settings.JOINT + self.parcel


class OperationDashboard:
    def __init__(self):
        main_dashboard = Dashboard(Operation)

        self.total_number = main_dashboard.total_number

        self.filters_keys = [
            'recorded', 'effective', 'active', 'field',
            'documented', 'closed', 'documented_closed']
        filters = {
            'recorded': {},
            'effective': {'scientist__isnull': False},
            'active': {'scientist__isnull': False, 'end_date__isnull': True},
            'field': {'excavation_end_date__isnull': True},
            'documented': {'source__isnull': False},
            'documented_closed': {'source__isnull': False,
                                  'end_date__isnull': False},
            'closed': {'end_date__isnull': False}
        }
        filters_label = {
            'recorded': _(u"Recorded"),
            'effective': _(u"Effective"),
            'active': _(u"Active"),
            'field': _(u"Field completed"),
            'documented': _(u"Associated report"),
            'closed': _(u"Closed"),
            'documented_closed': _(u"Documented and closed"),
        }
        self.filters_label = [filters_label[k] for k in self.filters_keys]
        self.total = []
        for fltr_key in self.filters_keys:
            fltr, lbl = filters[fltr_key], filters_label[fltr_key]
            nb = Operation.objects.filter(**fltr).count()
            self.total.append((lbl, nb))

        self.surface_by_type = Operation.objects\
            .values('operation_type__label')\
            .annotate(number=Sum('surface'))\
            .order_by('-number', 'operation_type__label')

        self.by_type = []
        self.types = OperationType.objects.filter(available=True).all()
        for fltr_key in self.filters_keys:
            fltr, lbl = filters[fltr_key], filters_label[fltr_key]
            type_res = Operation.objects.filter(**fltr).\
                values('operation_type', 'operation_type__label').\
                annotate(number=Count('pk')).\
                order_by('operation_type')
            types_dct = {}
            for typ in type_res.all():
                types_dct[typ['operation_type']] = typ["number"]
            types = []
            for typ in self.types:
                if typ.pk in types_dct:
                    types.append(types_dct[typ.pk])
                else:
                    types.append(0)
            self.by_type.append((lbl, types))

        self.by_year = []
        self.years = [res['year'] for res in Operation.objects.values('year')
                      .order_by('-year').distinct()]
        for fltr_key in self.filters_keys:
            fltr, lbl = filters[fltr_key], filters_label[fltr_key]
            year_res = Operation.objects.filter(**fltr)\
                                        .values('year')\
                                        .annotate(number=Count('pk'))\
                                        .order_by('year')
            years_dct = {}
            for yr in year_res.all():
                years_dct[yr['year']] = yr["number"]
            years = []
            for yr in self.years:
                if yr in years_dct:
                    years.append(years_dct[yr])
                else:
                    years.append(0)
            self.by_year.append((lbl, years))

        self.by_realisation_year = []
        self.realisation_years = [
            res['date'] for res in Operation.objects.extra(
                {'date': "date_trunc('year', start_date)"}).values('date')
            .filter(start_date__isnull=False).order_by('-date').distinct()]
        for fltr_key in self.filters_keys:
            fltr, lbl = filters[fltr_key], filters_label[fltr_key]
            year_res = Operation.objects.filter(**fltr).extra(
                {'date': "date_trunc('year', start_date)"}).values('date')\
                .values('date').filter(start_date__isnull=False)\
                .annotate(number=Count('pk'))\
                .order_by('-date')
            years_dct = {}
            for yr in year_res.all():
                years_dct[yr['date']] = yr["number"]
            years = []
            for yr in self.realisation_years:
                if yr in years_dct:
                    years.append(years_dct[yr])
                else:
                    years.append(0)
            self.by_realisation_year.append((lbl, years))

        self.effective = []
        for typ in self.types:
            year_res = Operation.objects.filter(**{'scientist__isnull': False,
                                                   'operation_type': typ})\
                                        .values('year')\
                                        .annotate(number=Count('pk'))\
                                        .order_by('-year').distinct()
            years_dct = {}
            for yr in year_res.all():
                years_dct[yr['year']] = yr["number"]
            years = []
            for yr in self.years:
                if yr in years_dct:
                    years.append(years_dct[yr])
                else:
                    years.append(0)
            self.effective.append((typ, years))

        # TODO: by date
        now = datetime.date.today()
        limit = datetime.date(now.year, now.month, 1) - datetime.timedelta(365)
        by_realisation_month = Operation.objects.filter(
            start_date__gt=limit, start_date__isnull=False).extra(
                {'date': "date_trunc('month', start_date)"})
        self.last_months = []
        date = datetime.datetime(now.year, now.month, 1)
        for mt_idx in xrange(12):
            self.last_months.append(date)
            if date.month > 1:
                date = datetime.datetime(date.year, date.month - 1, 1)
            else:
                date = datetime.datetime(date.year - 1, 12, 1)
        self.by_realisation_month = []
        for fltr_key in self.filters_keys:
            fltr, lbl = filters[fltr_key], filters_label[fltr_key]
            month_res = by_realisation_month.filter(**fltr)\
                .annotate(number=Count('pk')).order_by('-date')
            month_dct = {}
            for mt in month_res.all():
                month_dct[mt.date] = mt.number
            date = datetime.date(now.year, now.month, 1)
            months = []
            for date in self.last_months:
                if date in month_dct:
                    months.append(month_dct[date])
                else:
                    months.append(0)
            self.by_realisation_month.append((lbl, months))

        # survey and excavations
        self.survey, self.excavation = {}, {}
        for dct_res, ope_types in ((self.survey, ('arch_diagnostic',)),
                                   (self.excavation, ('prev_excavation',
                                                      'prog_excavation'))):
            dct_res['total'] = []
            operation_type = {'operation_type__txt_idx__in': ope_types}
            for fltr_key in self.filters_keys:
                fltr, lbl = filters[fltr_key], filters_label[fltr_key]
                fltr.update(operation_type)
                nb = Operation.objects.filter(**fltr).count()
                dct_res['total'].append((lbl, nb))

            dct_res['by_year'] = []
            for fltr_key in self.filters_keys:
                fltr, lbl = filters[fltr_key], filters_label[fltr_key]
                fltr.update(operation_type)
                year_res = Operation.objects.filter(**fltr)\
                                            .values('year')\
                                            .annotate(number=Count('pk'))\
                                            .order_by('year')
                years_dct = {}
                for yr in year_res.all():
                    years_dct[yr['year']] = yr["number"]
                years = []
                for yr in self.years:
                    if yr in years_dct:
                        years.append(years_dct[yr])
                    else:
                        years.append(0)
                dct_res['by_year'].append((lbl, years))

            dct_res['by_realisation_year'] = []
            for fltr_key in self.filters_keys:
                fltr, lbl = filters[fltr_key], filters_label[fltr_key]
                fltr.update(operation_type)
                year_res = Operation.objects.filter(**fltr).extra(
                    {'date': "date_trunc('year', start_date)"})\
                    .values('date')\
                    .filter(start_date__isnull=False)\
                    .annotate(number=Count('pk'))\
                    .order_by('-date')
                years_dct = {}
                for yr in year_res.all():
                    years_dct[yr['date']] = yr["number"]
                years = []
                for yr in self.realisation_years:
                    if yr in years_dct:
                        years.append(years_dct[yr])
                    else:
                        years.append(0)
                dct_res['by_realisation_year'].append((lbl, years))

            current_year_ope = Operation.objects.filter(**operation_type)\
                                        .filter(
                                            year=datetime.date.today().year)
            current_realisation_year_ope = Operation.objects\
                .filter(**operation_type)\
                .filter(start_date__year=datetime.date.today().year)
            res_keys = [('area_realised', current_realisation_year_ope)]
            if dct_res == self.survey:
                res_keys.append(('area',
                                 current_year_ope))
            for res_key, base_ope in res_keys:
                dct_res[res_key] = []
                for fltr_key in self.filters_keys:
                    fltr, lbl = filters[fltr_key], filters_label[fltr_key]
                    area_res = base_ope.filter(**fltr)\
                        .annotate(number=Sum('surface')).all()
                    val = 0
                    if area_res:
                        val = (area_res[0].number or 0) / 10000.0
                    dct_res[res_key].append(val)
            # TODO...
            res_keys = [('manday_realised', current_realisation_year_ope)]
            if dct_res == self.survey:
                res_keys.append(('manday',
                                 current_year_ope))
            for res_key, base_ope in res_keys:
                dct_res[res_key] = []
                for fltr_key in self.filters_keys:
                    dct_res[res_key].append('-')
            # TODO...
            res_keys = [('mandayhect_realised', current_realisation_year_ope)]
            if dct_res == self.survey:
                res_keys.append(('mandayhect',
                                 current_year_ope))
            for res_key, base_ope in res_keys:
                dct_res[res_key] = []
                for fltr_key in self.filters_keys:
                    dct_res[res_key].append('-')
            # TODO...
            dct_res['mandayhect_real_effective'] = '-'
            if dct_res == self.survey:
                dct_res['mandayhect_effective'] = '-'

            res_keys = [('org_realised', current_realisation_year_ope)]
            if dct_res == self.survey:
                res_keys.append(('org', current_year_ope))
            for res_key, base_ope in res_keys:
                org_res = base_ope.filter(scientist__attached_to__isnull=False)\
                    .values('scientist__attached_to',
                            'scientist__attached_to__name')\
                    .annotate(area=Sum('surface'))\
                    .order_by('scientist__attached_to__name').all()
                # TODO: man-days, man-days/hectare

                dct_res[res_key] = []
                for vals in org_res:
                    vals['area'] = (vals['area'] or 0) / 10000.0
                    dct_res[res_key].append(vals)

            year_ope = Operation.objects.filter(**operation_type)
            res_keys = ['org_by_year']
            if dct_res == self.survey:
                res_keys.append('org_by_year_realised')
            q = year_ope.values('scientist__attached_to',
                                'scientist__attached_to__name')\
                .filter(scientist__attached_to__isnull=False)\
                .order_by('scientist__attached_to__name').distinct()
            org_list = [(org['scientist__attached_to'],
                         org['scientist__attached_to__name']) for org in q]
            # org_list_dct = dict(org_list)
            for res_key in res_keys:
                dct_res[res_key] = []
                years = self.years
                if res_key == 'org_by_year_realised':
                    years = self.realisation_years
                for org_id, org_label in org_list:
                    org_res = year_ope.filter(
                        scientist__attached_to__pk=org_id)
                    key_date = ''
                    if res_key == 'org_by_year':
                        org_res = org_res.values('year')
                        key_date = 'year'
                    else:
                        org_res = org_res\
                            .extra({'date': "date_trunc('year', start_date)"})\
                            .values('date')\
                            .filter(start_date__isnull=False)
                        key_date = 'date'
                    org_res = org_res.annotate(area=Sum('surface'),
                                               cost=Sum('cost'))
                    years_dct = {}
                    for yr in org_res.all():
                        area = (yr['area'] if yr['area'] else 0) / 10000.0
                        cost = yr['cost'] if yr['cost'] else 0
                        years_dct[yr[key_date]] = (area, cost)
                    r_years = []
                    for yr in years:
                        if yr in years_dct:
                            r_years.append(years_dct[yr])
                        else:
                            r_years.append((0, 0))
                    dct_res[res_key].append((org_label, r_years))
                area_means, area_sums = [], []
                cost_means, cost_sums = [], []
                for idx, year in enumerate(years):
                    vals = [r_yars[idx] for lb, r_yars in dct_res[res_key]]
                    if not vals:
                        continue
                    sum_area = sum([a for a, c in vals])
                    sum_cost = sum([c for a, c in vals])
                    area_means.append(sum_area / len(vals))
                    area_sums.append(sum_area)
                    cost_means.append(sum_cost / len(vals))
                    cost_sums.append(sum_cost)
                dct_res[res_key + '_area_mean'] = area_means
                dct_res[res_key + '_area_sum'] = area_sums
                dct_res[res_key + '_cost_mean'] = cost_means
                dct_res[res_key + '_cost_mean'] = cost_sums

            if dct_res == self.survey:
                self.survey['effective'] = []
                for yr in self.years:
                    year_res = Operation.objects\
                        .filter(scientist__isnull=False, year=yr,
                                operation_type__txt_idx__in=ope_types)\
                        .annotate(number=Sum('surface'), mean=Avg('surface'))
                    nb = year_res[0].number if year_res.count() else 0
                    nb = nb if nb else 0
                    mean = year_res[0].mean if year_res.count() else 0
                    mean = mean if mean else 0
                    self.survey['effective'].append((nb, mean))

            # TODO:Man-Days/hectare by Year

            # CHECK: month of realisation or month?
            dct_res['by_month'] = []
            for fltr_key in self.filters_keys:
                fltr, lbl = filters[fltr_key], filters_label[fltr_key]
                fltr.update(operation_type)
                month_res = by_realisation_month\
                    .filter(**fltr)\
                    .annotate(number=Count('pk'))\
                    .order_by('-date')
                month_dct = {}
                for mt in month_res.all():
                    month_dct[mt.date] = mt.number
                date = datetime.date(now.year, now.month, 1)
                months = []
                for date in self.last_months:
                    if date in month_dct:
                        months.append(month_dct[date])
                    else:
                        months.append(0)
                dct_res['by_month'].append((lbl, months))

            operation_type = {'operation_type__txt_idx__in': ope_types}
            self.departments = [
                (fd['department__pk'], fd['department__label'])
                for fd in OperationByDepartment
                .objects.filter(department__isnull=False)
                .values('department__label', 'department__pk')
                .order_by('department__label').distinct()]
            dct_res['by_dpt'] = []
            for dpt_id, dpt_label in self.departments:
                vals = OperationByDepartment.objects\
                    .filter(department__pk=dpt_id,
                            operation__operation_type__txt_idx__in=ope_types)\
                    .values('department__pk', 'operation__year')\
                    .annotate(number=Count('operation'))\
                    .order_by('operation__year')
                dct_years = {}
                for v in vals:
                    dct_years[v['operation__year']] = v['number']
                years = []
                for y in self.years:
                    if y in dct_years:
                        years.append(dct_years[y])
                    else:
                        years.append(0)
                years.append(sum(years))
                dct_res['by_dpt'].append((dpt_label, years))
            dct_res['effective_by_dpt'] = []
            for dpt_id, dpt_label in self.departments:
                vals = OperationByDepartment.objects\
                    .filter(department__pk=dpt_id,
                            operation__scientist__isnull=False,
                            operation__operation_type__txt_idx__in=ope_types)\
                    .values('department__pk', 'operation__year')\
                    .annotate(number=Count('operation'),
                              area=Sum('operation__surface'),
                              fnap=Sum('operation__fnap_cost'),
                              cost=Sum('operation__cost'))\
                    .order_by('operation__year')
                dct_years = {}
                for v in vals:
                    values = []
                    for k in ('number', 'area', 'cost', 'fnap'):
                        value = v[k] or 0
                        if k == 'area':
                            value /= 10000.0
                        values.append(value)

                    dct_years[v['operation__year']] = values
                years = []
                for y in self.years:
                    if y in dct_years:
                        years.append(dct_years[y])
                    else:
                        years.append((0, 0, 0, 0))
                nbs, areas, costs, fnaps = zip(*years)
                years.append((sum(nbs), sum(areas), sum(costs), sum(fnaps)))
                dct_res['effective_by_dpt'].append((dpt_label, years))

            OperationTown = Operation.towns.through
            query = OperationTown.objects\
                .filter(operation__scientist__isnull=False,
                        operation__operation_type__txt_idx__in=ope_types)\
                .values('town__name', 'town__departement__number')\
                .annotate(nb=Count('operation'))\
                .order_by('-nb', 'town__name')[:10]
            dct_res['towns'] = []
            for r in query:
                dct_res['towns'].append((u"%s (%s)" % (r['town__name'],
                                         r['town__departement__number']),
                                         r['nb']))

            if dct_res == self.survey:
                query = OperationTown.objects\
                    .filter(operation__scientist__isnull=False,
                            operation__operation_type__txt_idx__in=ope_types,
                            operation__surface__isnull=False)\
                    .values('town__name', 'town__departement__number')\
                    .annotate(nb=Sum('operation__surface'))\
                    .order_by('-nb', 'town__name')[:10]
                dct_res['towns_surface'] = []
                for r in query:
                    dct_res['towns_surface'].append((u"%s (%s)" % (
                        r['town__name'], r['town__departement__number']),
                        r['nb']))
            else:
                query = OperationTown.objects\
                    .filter(operation__scientist__isnull=False,
                            operation__operation_type__txt_idx__in=ope_types,
                            operation__cost__isnull=False)\
                    .values('town__name', 'town__departement__number')\
                    .annotate(nb=Sum('operation__cost'))\
                    .order_by('-nb', 'town__name')[:10]
                dct_res['towns_cost'] = []
                for r in query:
                    dct_res['towns_cost'].append((u"%s (%s)" % (
                        r['town__name'], r['town__departement__number']),
                        r['nb']))


class OperationTypeOld(GeneralType):
    order = models.IntegerField(_(u"Order"), default=1)
    preventive = models.BooleanField(_(u"Is preventive"), default=True)

    class Meta:
        verbose_name = _(u"Operation type old")
        verbose_name_plural = _(u"Operation types old")
        ordering = ['-preventive', 'order', 'label']



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2012-2017 Étienne Loks  <etienne.loks_AT_peacefrogsDOTnet>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# See the file COPYING for details.

import json
import datetime

from django.conf import settings
from django.core.files.uploadedfile import SimpleUploadedFile
from django.core.urlresolvers import reverse
from django.test.client import Client

from django.contrib.auth.models import Permission
import models

from archaeological_operations import views

from ishtar_common.models import OrganizationType, Organization, ItemKey, \
    ImporterType, IshtarUser, TargetKey, ImporterModel, IshtarSiteProfile, \
    Town, ImporterColumn, Person
from archaeological_context_records.models import Unit

from ishtar_common import forms_common
from ishtar_common.tests import WizardTest, WizardTestFormData as FormData, \
    create_superuser, create_user, TestCase


class ImportTest(object):
    def setUp(self):
        self.username, self.password, self.user = create_superuser()
        self.ishtar_user = IshtarUser.objects.get(pk=self.user.pk)

    def set_target_key(self, target, key, value, imp=None):
        keys = {'target__target': target, 'key': key}
        if imp:
            keys['associated_import'] = imp
        tg = TargetKey.objects.get(**keys)
        tg.value = value
        tg.is_set = True
        tg.save()

    def init_ope_import(self, filename='MCC-operations-example.csv'):
        mcc_operation = ImporterType.objects.get(name=u"MCC - Opérations")
        mcc_operation_file = open(
            settings.ROOT_PATH +
            '../archaeological_operations/tests/' + filename,
            'rb')
        file_dict = {'imported_file': SimpleUploadedFile(
            mcc_operation_file.name, mcc_operation_file.read())}
        post_dict = {'importer_type': mcc_operation.pk, 'skip_lines': 1,
                     "encoding": 'utf-8'}
        form = forms_common.NewImportForm(data=post_dict, files=file_dict)
        form.is_valid()
        return mcc_operation, form

    def init_ope_targetkey(self, imp):
        # doing manually connections
        target = TargetKey.objects.filter(
            target__target='operation_type').order_by('-pk').all()[0]
        target.value = models.OperationType.objects.get(
            txt_idx='prog_excavation').pk
        target.is_set = True
        target.associated_import = imp
        target.save()

        target2 = TargetKey.objects.get(key='gallo-romain',
                                        associated_import=imp)
        gallo = models.Period.objects.get(txt_idx='gallo-roman')
        target2.value = gallo.pk
        target2.is_set = True
        target2.associated_import = imp
        target2.save()

        target3 = TargetKey.objects.get(key='age-du-fer',
                                        associated_import=imp)
        iron = models.Period.objects.get(txt_idx='iron_age')
        target3.value = iron.pk
        target3.is_set = True
        target3.associated_import = imp
        target3.save()
        return [target, target2, target3]

    def init_ope(self):
        importer, form = self.init_ope_import()
        impt = form.save(self.ishtar_user)
        impt.initialize()
        self.init_ope_targetkey(imp=impt)
        impt.importation()

    def init_parcel_import(self):
        self.init_ope()
        mcc_parcel = ImporterType.objects.get(name=u"MCC - Parcelles")
        mcc_file = open(
            settings.ROOT_PATH +
            '../archaeological_operations/tests/MCC-parcelles-example.csv',
            'rb')
        file_dict = {'imported_file': SimpleUploadedFile(mcc_file.name,
                                                         mcc_file.read())}
        post_dict = {'importer_type': mcc_parcel.pk, 'skip_lines': 1,
                     "encoding": 'utf-8'}
        form = forms_common.NewImportForm(data=post_dict, files=file_dict)
        form.is_valid()
        return mcc_parcel, form

    def init_parcel(self):
        importer, form = self.init_parcel_import()
        impt = form.save(self.ishtar_user)
        impt.initialize()
        impt.importation()

    def init_context_record_import(self):
        self.init_parcel()
        mcc = ImporterType.objects.get(name=u"MCC - UE")
        mcc_file = open(
            settings.ROOT_PATH +
            '../archaeological_context_records/tests/'
            'MCC-context-records-example.csv', 'rb')
        file_dict = {'imported_file': SimpleUploadedFile(mcc_file.name,
                                                         mcc_file.read())}
        post_dict = {'importer_type': mcc.pk, 'skip_lines': 1,
                     "encoding": 'utf-8'}
        form = forms_common.NewImportForm(data=post_dict, files=file_dict)
        form.is_valid()
        return mcc, form

    def init_cr_targetkey(self, imp):
        hc = Unit.objects.get(txt_idx='not_in_context').pk
        self.set_target_key('unit', 'hc', hc, imp=imp)
        self.set_target_key('unit', 'hors-contexte', hc, imp=imp)
        layer = Unit.objects.get(txt_idx='negative').pk
        self.set_target_key('unit', 'couche', layer, imp=imp)

    def init_context_record(self):
        mcc, form = self.init_context_record_import()
        impt = form.save(self.ishtar_user)
        impt.initialize()
        self.init_cr_targetkey(impt)
        impt.importation()


class ImportOperationTest(ImportTest, TestCase):
    fixtures = [settings.ROOT_PATH +
                '../fixtures/initial_data-auth-fr.json',
                settings.ROOT_PATH +
                '../ishtar_common/fixtures/initial_data-fr.json',
                settings.ROOT_PATH +
                '../ishtar_common/fixtures/test_towns.json',
                settings.ROOT_PATH +
                '../ishtar_common/fixtures/initial_importtypes-fr.json',
                settings.ROOT_PATH +
                '../archaeological_operations/fixtures/initial_data-fr.json']

    def test_mcc_import_operation(self):
        first_ope_nb = models.Operation.objects.count()
        first_person_nb = Person.objects.count()
        importer, form = self.init_ope_import()
        self.assertTrue(form.is_valid())
        impt = form.save(self.ishtar_user)
        target_key_nb = TargetKey.objects.count()
        impt.initialize()
        # new key have to be set
        self.assertTrue(TargetKey.objects.count() > target_key_nb)

        # first try to import
        impt.importation()
        current_ope_nb = models.Operation.objects.count()
        # no new operation imported because of a missing connection for
        # operation_type value
        self.assertEqual(current_ope_nb, first_ope_nb)
        self.init_ope_targetkey(imp=impt)

        impt.importation()
        # new operations have now been imported
        current_ope_nb = models.Operation.objects.count()
        self.assertEqual(current_ope_nb, first_ope_nb + 2)
        current_person_nb = Person.objects.count()
        self.assertEqual(current_person_nb, first_person_nb + 1)
        # and well imported
        last_ope = models.Operation.objects.order_by('-pk').all()[0]
        self.assertEqual(last_ope.name, u"Oppìdum de Paris")
        self.assertEqual(last_ope.code_patriarche, 4200)
        self.assertEqual(last_ope.operation_type.txt_idx, 'prog_excavation')
        self.assertEqual(last_ope.periods.count(), 2)
        periods = [period.txt_idx for period in last_ope.periods.all()]
        self.assertIn('iron_age', periods)
        self.assertIn('gallo-roman', periods)

        # a second importation will be not possible: no two same patriarche
        # code
        impt.importation()
        self.assertEqual(last_ope,
                         models.Operation.objects.order_by('-pk').all()[0])

    def test_import_bad_encoding(self):
        self.init_ope_import('MCC-operations-example-bad-encoding.csv')

    def test_keys_limitation(self):
        # each key association is associated to the import
        init_ope_number = models.Operation.objects.count()
        importer, form = self.init_ope_import()
        impt = form.save(self.ishtar_user)
        impt.initialize()
        self.init_ope_targetkey(imp=impt)

        importer, form = self.init_ope_import()
        other_imp = form.save(self.ishtar_user)
        # associate with another import
        for ik in ItemKey.objects.filter(importer=impt).all():
            ik.importer = other_imp
            ik.save()

        impt.importation()
        current_ope_nb = models.Operation.objects.count()
        # no new operation
        self.assertEqual(current_ope_nb, init_ope_number)

    def test_bad_configuration(self):
        importer, form = self.init_ope_import()
        col = ImporterColumn.objects.get(importer_type=importer, col_number=1)
        target = col.targets.all()[0]
        target.target = "cody"  # random and not appropriate string
        target.save()
        # self.init_ope()
        # importer, form = self.init_ope_import()
        impt = form.save(self.ishtar_user)
        impt.initialize()
        self.init_ope_targetkey(imp=impt)
        impt.importation()
        self.assertEqual(len(impt.errors), 2)
        self.assertTrue(
            "Importer configuration error" in impt.errors[0]['error'] or
            "Erreur de configuration de l\'importeur" in impt.errors[0]['error']
        )

    def test_model_limitation(self):
        importer, form = self.init_ope_import()
        importer.created_models.clear()
        impt = form.save(self.ishtar_user)
        impt.initialize()
        self.init_ope_targetkey(imp=impt)

        # no model defined in created_models: normal import
        init_ope_number = models.Operation.objects.count()
        impt.importation()
        current_ope_nb = models.Operation.objects.count()
        self.assertEqual(current_ope_nb, init_ope_number + 2)

        for ope in models.Operation.objects.order_by('-pk').all()[:2]:
            ope.delete()

        importer, form = self.init_ope_import()
        # add an inadequate model to make created_models non empty
        importer.created_models.clear()
        importer.created_models.add(ImporterModel.objects.get(
            klass='ishtar_common.models.Organization'
        ))
        impt = form.save(self.ishtar_user)
        impt.initialize()
        self.init_ope_targetkey(imp=impt)

        # no imports
        impt.importation()
        current_ope_nb = models.Operation.objects.count()
        self.assertEqual(current_ope_nb, init_ope_number)

        importer, form = self.init_ope_import()
        # add operation model to allow creation
        importer.created_models.clear()
        importer.created_models.add(ImporterModel.objects.get(
            klass='archaeological_operations.models.Operation'
        ))
        impt = form.save(self.ishtar_user)
        impt.initialize()
        self.init_ope_targetkey(imp=impt)

        # import of operations
        impt.importation()
        current_ope_nb = models.Operation.objects.count()
        self.assertEqual(current_ope_nb, init_ope_number + 2)

    def test_mcc_import_parcels(self):
        old_nb = models.Parcel.objects.count()
        mcc_parcel, form = self.init_parcel_import()
        impt = form.save(self.ishtar_user)
        impt.initialize()
        impt.importation()
        # new parcels has now been imported
        current_nb = models.Parcel.objects.count()
        self.assertEqual(current_nb, old_nb + 3)

        # and well imported
        last_parcels = models.Parcel.objects.order_by('-pk').all()[0:3]
        external_ids = sorted(['4200-59350-YY55', '4200-75101-XXXX',
                               '4201-59350-YY55'])
        parcel_numbers = sorted(['42', '55', '55'])
        sections = sorted(['ZX', 'YY', 'YY'])
        self.assertEqual(external_ids,
                         sorted([p.external_id for p in last_parcels]))
        self.assertEqual(parcel_numbers,
                         sorted([p.parcel_number for p in last_parcels]))
        self.assertEqual(sections,
                         sorted([p.section for p in last_parcels]))
        ope1 = models.Operation.objects.filter(code_patriarche=4200).all()[0]
        towns_ope = ope1.towns.all()
        imported = [imp for acc, imp in impt.get_all_imported()]
        for p in last_parcels:
            self.assertTrue(p.town in towns_ope)
            self.assertTrue(p in imported)
        self.assertEqual(len(imported), len(last_parcels))
        self.assertEqual(
            models.Parcel.objects.get(
                parcel_number='55', section='YY',
                operation_id=ope1.pk).external_id,
            '4200-59350-YY55')
        # cached_label update
        ope2 = models.Operation.objects.filter(code_patriarche=4201).all()[0]
        self.assertIn('LILLE', ope2.cached_label.upper())
        # delete associated parcel with the import deletion
        parcel_count = models.Parcel.objects.count()
        impt.delete()
        self.assertEqual(parcel_count - 3, models.Parcel.objects.count())


class ParcelTest(ImportTest, TestCase):
    fixtures = [settings.ROOT_PATH +
                '../fixtures/initial_data-auth-fr.json',
                settings.ROOT_PATH +
                '../ishtar_common/fixtures/initial_data-fr.json',
                settings.ROOT_PATH +
                '../ishtar_common/fixtures/test_towns.json',
                settings.ROOT_PATH +
                '../ishtar_common/fixtures/initial_importtypes-fr.json',
                settings.ROOT_PATH +
                '../archaeological_operations/fixtures/initial_data-fr.json']

    def test_parse_parcels(self):
        # the database needs to be initialised before importing
        from archaeological_operations.import_from_csv import parse_parcels
        # default_town = Town.objects.create(numero_insee="12345",
        #                                    name="default_town")
        test_values = (
            (u"1996 : XT:53,54,56,57,59,60,61,62",
             {1996: [
                 ("XT", "53"), ("XT", "54"), ("XT", "56"), ("XT", "57"),
                 ("XT", "59"), ("XT", "60"), ("XT", "61"), ("XT", "62"),
             ]}
             ),
            (u"AD:23",
             {None: [
                 ("AD", "23")
             ]}),
            (u"1961 :B1:227;",
             {1961: [
                 ("B1", '227')
             ]}),
            (u"1982 CV:35;CV:36",
             {1982: [
                 ("CV", "35"), ("CV", "36"),
             ]}),
            (u"E:24;E:25",
             {None: [
                 ("E", "24"), ("E", "25"),
             ]}),
            (u"B : 375, 376, 386, 387, 645, 646 / C : 412 à 415, 432 à 435, "
             u"622 / F : 120, 149, 150, 284, 287, 321 à 323",
             {None: [
                 ("B", "375"), ("B", "376"), ("B", "386"), ("B", "387"),
                 ("B", "645"), ("B", "646"),
                 ("C", "412"), ("C", "413"), ("C", "414"), ("C", "415"),
                 ("C", "432"), ("C", "433"), ("C", "434"), ("C", "435"),
                 ("C", "622"),
                 ("F", "120"), ("F", "149"), ("F", "150"), ("F", "284"),
                 ("F", "287"), ("F", "321"), ("F", "322"), ("F", "323"),
             ]}),
            (u"AD : 95, 96, 86, 87, 81, 252, AE : 58, AD : 115 à 132",
             {None: [
                 ("AD", "95"), ("AD", "96"), ("AD", "86"), ("AD", "87"),
                 ("AD", "81"), ("AD", "252"), ("AD", "115"), ("AD", "116"),
                 ("AD", "117"), ("AD", "118"), ("AD", "119"), ("AD", "120"),
                 ("AD", "121"), ("AD", "122"), ("AD", "123"), ("AD", "124"),
                 ("AD", "125"), ("AD", "126"), ("AD", "127"), ("AD", "128"),
                 ("AD", "129"), ("AD", "130"), ("AD", "131"), ("AD", "132"),
                 ("AE", "58"),
             ]}),
            (u"XD:1 à 13, 24 à 28, 33 à 39, 50 à 52, 80, 83, 84 à 86, 259 à "
             u"261, 182, 225 ; XH:5 ; P:1640, 1888, 1889, 1890 ; R:1311, "
             u"1312, 1314,1342, 1343, 1559 à 1569",
             {None: [
                 ('XD', "1"), ('XD', "2"), ('XD', "3"), ('XD', "4"),
                 ('XD', "5"), ('XD', "6"), ('XD', "7"), ('XD', "8"),
                 ('XD', "9"), ('XD', "10"), ('XD', "11"), ('XD', "12"),
                 ('XD', "13"), ("XD", "24"), ("XD", "25"), ("XD", "26"),
                 ("XD", "27"), ("XD", "28"), ("XD", "33"), ("XD", "34"),
                 ("XD", "35"), ("XD", "36"), ("XD", "37"), ("XD", "38"),
                 ("XD", "39"), ("XD", "50"), ("XD", "51"), ("XD", "52"),
                 ("XD", "80"), ("XD", "83"), ("XD", "84"), ("XD", "85"),
                 ("XD", "86"), ("XD", "259"), ("XD", "260"), ("XD", "261"),
                 ("XD", "182"), ("XD", "225"), ("XH", "5"),
                 ("P", "1640"), ("P", "1888"), ("P", "1889"), ("P", "1890"),
                 ("R", "1311"), ("R", "1312"), ("R", "1314"), ("R", "1342"),
                 ("R", "1343"), ("R", "1559"), ("R", "1560"), ("R", "1561"),
                 ("R", "1562"), ("R", "1563"), ("R", "1564"), ("R", "1565"),
                 ("R", "1566"), ("R", "1567"), ("R", "1568"), ("R", "1569"),
             ]}),
            (u"BZ:2 à 5, 365 ; CD:88 à 104, 106, 108, 326",
             {None: [
                 ('BZ', '2'), ('BZ', '3'), ('BZ', '4'), ('BZ', '5'),
                 ('BZ', '365'), ('CD', '88'), ('CD', '89'), ('CD', '90'),
                 ('CD', '91'), ('CD', '92'), ('CD', '93'), ('CD', '94'),
                 ('CD', '95'), ('CD', '96'), ('CD', '97'), ('CD', '98'),
                 ('CD', '99'), ('CD', '100'), ('CD', '101'), ('CD', '102'),
                 ('CD', '103'), ('CD', '104'), ('CD', '106'), ('CD', '326'),
                 ('CD', '108')
             ]}),
            (u"AV 118 à 125, 127, 132 à 137, 153, 398p, 399, 402; BI 27, 30, "
             u"32, 33, 188, 255, 256 à 258, 260, 284p, 294; BL 297",
             {None: [
                 ('AV', '118'), ('AV', '119'), ('AV', '120'), ('AV', '121'),
                 ('AV', '122'), ('AV', '123'), ('AV', '124'), ('AV', '125'),
                 ('AV', '127'), ('AV', '132'), ('AV', '133'), ('AV', '134'),
                 ('AV', '135'), ('AV', '136'), ('AV', '137'), ('AV', '153'),
                 ('AV', '398p'), ('AV', '399'), ('AV', '402'),
                 ('BI', '27'), ('BI', '30'), ('BI', '32'), ('BI', '33'),
                 ('BI', '188'), ('BI', '255'), ('BI', '256'), ('BI', '257'),
                 ('BI', '258'), ('BI', '260'), ('BI', '284p'), ('BI', '294'),
                 ('BL', '297'),
             ]}),
            (u"A : 904 à 906, 911 ; E:40, 41",
             {None: [
                 ("A", '904'), ("A", '905'), ("A", '906'), ("A", '911'),
                 ("E", '40'), ("E", "41")
             ]}),
            (u"1991 : BE:8, 12",
             {"1991": [
                 ('BE', '8'), ('BE', '12'),
             ]}),
            (u"1979 : EM:1",
             {"1979": [
                 ('EM', '1')
             ]},),
            (u"B:448;B:449;B:450;B:451;B:452;B:455;B:456;B:457;B:458;B:459;"
             u"B:1486;",
             {None: [
                 ("B", "448"), ("B", "449"), ("B", "450"), ("B", "451"),
                 ("B", "452"), ("B", "455"), ("B", "456"), ("B", "457"),
                 ("B", "458"), ("B", "459"), ("B", "1486"),
             ]}),
            (u"AC : 72 à 81, 91 à 100, 197 / ZC:180 à 189",
             {None: [
                 ('AC', '72'), ('AC', '73'), ('AC', '74'), ('AC', '75'),
                 ('AC', '76'), ('AC', '77'), ('AC', '78'), ('AC', '79'),
                 ('AC', '80'), ('AC', '81'), ('AC', '91'), ('AC', '92'),
                 ('AC', '93'), ('AC', '94'), ('AC', '95'), ('AC', '96'),
                 ('AC', '97'), ('AC', '98'), ('AC', '99'), ('AC', '100'),
                 ('AC', '197'), ('ZC', '180'), ('ZC', '181'), ('ZC', '182'),
                 ('ZC', '183'), ('ZC', '184'), ('ZC', '185'), ('ZC', '186'),
                 ('ZC', '187'), ('ZC', '188'), ('ZC', '189'),
             ]}),
            (u"AB 37 et 308",
             {None: [
                 ('AB', '37'), ('AB', '308'),
             ]}),
            (u"1983  D2 n° 458 et 459",
             {"1983": [
                 ('D2', '458'), ('D2', '459'),
             ]}),
            (u"ZS : 21p, 66",
             {None: [
              ('ZS', '21p'), ('ZS', '66'),
              ]}),
            (u"VV:166, 167, domaine public",
             {None: [
                 ('VV', '166'), ('VV', '167'),
             ]}),
            (u" AS:13 à 15, 17 à 19, 21 à 32, 34 à 45, 47 à 53, 69, 70, 82, "
             u"84 / CK:1, 24, 25, 29, 30, 37 à 43",
             {None: [
              ("AS", "13"), ("AS", "14"), ("AS", "15"), ("AS", "17"),
              ("AS", "18"), ("AS", "19"), ("AS", "21"), ("AS", "22"),
              ("AS", "23"), ("AS", "24"), ("AS", "25"), ("AS", "26"),
              ("AS", "27"), ("AS", "28"), ("AS", "29"), ("AS", "30"),
              ("AS", "31"), ("AS", "32"), ("AS", "34"), ("AS", "35"),
              ("AS", "36"), ("AS", "37"), ("AS", "38"), ("AS", "39"),
              ("AS", "40"), ("AS", "41"), ("AS", "42"), ("AS", "43"),
              ("AS", "44"), ("AS", "45"), ("AS", "47"), ("AS", "48"),
              ("AS", "49"), ("AS", "50"), ("AS", "51"), ("AS", "52"),
              ("AS", "53"), ('AS', "69"), ('AS', "70"), ('AS', "82"),
              ('AS', "84"), ('CK', "1"), ('CK', "24"), ('CK', "25"),
              ('CK', "29"), ('CK', "30"), ('CK', "37"), ('CK', "38"),
              ('CK', "39"), ('CK', "40"), ('CK', "41"), ('CK', "42"),
              ('CK', "43"), ]}),
            (u" ZN:37, 15, 35, 28, 29 / ZM:9, 73",
             {None: [
                 ("ZN", "37"), ("ZN", "15"), ("ZN", "35"), ("ZN", "28"),
                 ("ZN", "29"), ("ZM", "9"), ("ZM", "73"),
             ]}),
            (u" Tranche n°1 : YP:243, 12, 14 à 16, 18 à 26, DP / Tranche n°2 :"
             u"YP:17, 307, 27, 308, 44 à 46, 683, BM:1, 250, 488 à 492",
             {None: [
                 ('YP', '243'), ('YP', '12'), ('YP', '14'), ('YP', '15'),
                 ('YP', '16'), ('YP', '18'), ('YP', '19'), ('YP', '20'),
                 ('YP', '21'), ('YP', '22'), ('YP', '23'), ('YP', '24'),
                 ('YP', '25'), ('YP', '26'), ('YP', '17'), ('YP', '27'),
                 ('YP', '308'), ('YP', '44'), ('YP', '45'), ('YP', '46'),
                 ('YP', '683'), ('YP', '307'), ('BM', '1'), ('BM', '250'),
                 ('BM', '488'), ('BM', '489'), ('BM', '490'), ('BM', '491'),
                 ('BM', '492'),
             ]}),
            (u" H : 106, 156, 158",
             {None: [
                 ('H', '106'), ('H', '156'), ('H', '158'),
             ]}),
            (u"Section YO : parcelles n° 19; 20",
             {None: [
                 ('YO', '19'), ('YO', '20'),
             ]}),
            (u"1991 :AI:23;19;20;21;22;181;AM:116;214;215;233;235",
             {u"1991": [
                 (u"AI", "19"), (u"AI", "20"), (u"AI", "21"), (u"AI", "22"),
                 (u"AI", "23"), (u"AI", "181"),
                 (u"AM", "116"), (u"AM", "214"), (u"AM", "215"),
                 (u"AM", "233"), (u"AM", "235"),
             ]})
        )
        # ),(u"Domaine public", {}
        # ),(u"Tranche 1 : AV:4 à 6, 18, 80, 104 / partiellement : 5 et 18", {}
        # ),(u" 1987 : ZD: ?", {}
        # ),(u"A:26a, 26b, 27 / AB:95 / AK:4, 12, 20", {}
        for value, result in test_values:
            parcels = parse_parcels(value)
            if not parcels and not result:
                continue
            self.assertTrue(parcels != [],
                            msg="No parcel parsed for \"%s\"" % value)
            parcels_copy = parcels[:]
            for year in result.keys():
                for values in parcels_copy:
                    if values['year'] != year and \
                       values['year'] != unicode(year):
                        continue
                    self.assertTrue(
                        (values['section'], values['parcel_number'])
                        in result[year],
                        msg="Section - Parcel number: \"%s - %s\" is not "
                        "in \"%s\"" % (
                            values['section'], values['parcel_number'],
                            unicode(result[year])))
                    parcels.pop(parcels.index(values))
                    result[year].pop(result[year].index(
                        (values['section'], values['parcel_number'])))
            # all parcels have been imported
            self.assertEqual(parcels, [], msg="Parcel(s): \"%s\" haven't be "
                             "recognized in \"%s\"" % (str(parcels), value))
            not_imported = [data for data in result.values() if data]
            self.assertEqual(
                not_imported, [], msg="Parcel(s): \"%s\" haven't be "
                "recognized in \"%s\"" % (str(not_imported), value))


def create_orga(user):
    orga_type, created = OrganizationType.objects.get_or_create(
        txt_idx='operator')
    orga, created = Organization.objects.get_or_create(
        name='Operator', organization_type=orga_type, history_modifier=user)
    return orga


def create_operation(user, orga=None, values={}):
    dct = {'year': 2010, 'operation_type_id': 1,
           'history_modifier': user}
    dct.update(values)
    if orga:
        dct['operator'] = orga
    operation = models.Operation.objects.create(**dct)
    return operation


class OperationInitTest(object):
    def create_user(self):
        username, password, self.user = create_user()

    def get_default_user(self):
        if not hasattr(self, 'user') or not self.user:
            self.create_user()
        return self.user

    def create_orgas(self, user=None):
        if not user:
            user = self.get_default_user()
        self.orgas = [create_orga(user)]
        return self.orgas

    def get_default_orga(self, user=None):
        if not hasattr(self, 'orgas') or not self.orgas:
            self.create_orgas(user)
        return self.orgas[0]

    def create_towns(self, datas={}):
        default = {'numero_insee': '12345', 'name': 'default_town'}
        default.update(datas)
        town = models.Town.objects.create(**default)
        if not hasattr(self, 'towns') or not self.towns:
            self.towns = []
        self.towns.append(town)
        return self.towns

    def get_default_town(self):
        towns = getattr(self, 'towns', None)
        if not towns:
            self.create_towns()
        return self.towns[0]

    def create_parcel(self, data={}):
        default = {'town': self.get_default_town(),
                   'section': 'A', 'parcel_number': '1'}
        if not hasattr(self, 'operations'):
            self.create_operation()
        default['operation'] = self.operations[0]
        default.update(data)
        if not getattr(self, 'parcels', None):
            self.parcels = []
        self.parcels.append(models.Parcel.objects.create(**default))
        return self.parcels

    def get_default_parcel(self, force=False):
        if force:
            return self.create_parcel()[-1]
        parcel = self.create_parcel()[0]
        if models.Parcel.objects.filter(pk=parcel.pk).count():
            return parcel
        self.parcels.pop(0)
        return self.create_operation()[-1]

    def create_operation(self, user=None, orga=None):
        if not orga:
            self.get_default_orga(user)
        if not user:
            self.get_default_user()
        if not getattr(self, 'operations', None):
            self.operations = []
        self.operations.append(create_operation(user, orga))
        return self.operations

    def get_default_operation(self, force=False):
        if force:
            return self.create_operation()[-1]
        ope = self.create_operation()[0]
        if models.Operation.objects.filter(pk=ope.pk).count():
            return ope
        self.operations.pop(0)
        return self.create_operation()[-1]

    def tearDown(self):
        # cleanup for further test
        if hasattr(self, 'user'):
            self.user.delete()
            self.user = None
        # all try/except is necessary for bad migrations on master...
        # should be removed at the next big version
        if hasattr(self, 'operations'):
            for ope in self.operations:
                try:
                    ope.delete()
                except:
                    pass
            self.operations = []
        if hasattr(self, 'parcels'):
            for p in self.parcels:
                try:
                    p.delete()
                except:
                    pass
            self.parcels = []


class OperationTest(TestCase, OperationInitTest):
    fixtures = [settings.ROOT_PATH +
                '../fixtures/initial_data-auth-fr.json',
                settings.ROOT_PATH +
                '../ishtar_common/fixtures/initial_data-fr.json',
                settings.ROOT_PATH +
                '../archaeological_files/fixtures/initial_data.json',
                settings.ROOT_PATH +
                '../archaeological_operations/fixtures/initial_data-fr.json']

    def setUp(self):
        IshtarSiteProfile.objects.get_or_create(
            slug='default', active=True)
        self.username, self.password, self.user = create_superuser()
        self.alt_username, self.alt_password, self.alt_user = create_user()
        self.alt_user.user_permissions.add(Permission.objects.get(
            codename='view_own_operation'))
        self.orgas = self.create_orgas(self.user)
        self.operations = self.create_operation(self.user, self.orgas[0])
        self.operations += self.create_operation(self.alt_user, self.orgas[0])
        self.item = self.operations[0]

    def testExternalID(self):
        self.item.code_patriarche = '123456789'
        self.item.save()
        parcel = self.get_default_parcel()
        parcel.operation = self.item
        parcel.save()
        correct_ext_id = u"{}-{}-{}{}".format(
            self.item.code_patriarche, parcel.town.numero_insee,
            parcel.section, parcel.parcel_number)
        self.assertEqual(parcel.external_id, correct_ext_id)
        # auto has been previously set
        parcel.external_id = 'blabla'
        parcel.save()
        self.assertEqual(parcel.external_id, correct_ext_id)
        # deactivate auto
        parcel.auto_external_id = False
        parcel.external_id = 'blabla'
        parcel.save()
        self.assertEqual(parcel.external_id, 'blabla')

    def create_relations(self):
        rel1 = models.RelationType.objects.create(
            symmetrical=True, label='Include', txt_idx='include')
        rel2 = models.RelationType.objects.create(
            symmetrical=False, label='Included', txt_idx='included',
            inverse_relation=rel1)
        models.RecordRelations.objects.create(
            left_record=self.operations[0],
            right_record=self.operations[1],
            relation_type=rel1)
        return rel1, rel2

    def testPostDeleteRelations(self):
        self.create_relations()
        self.operations[0].delete()

    def testPostDeleteParcels(self):
        ope = self.operations[0]
        town = Town.objects.create(name='plouf', numero_insee='20000')
        parcel = models.Parcel.objects.create(town=town)
        parcel_nb = models.Parcel.objects.count()
        ope.parcels.add(parcel)
        ope.delete()
        # our parcel has no operation attached and should be deleted
        self.assertEqual(parcel_nb - 1, models.Parcel.objects.count())
        ope = self.operations[1]
        parcel = models.Parcel.objects.create(town=town)
        parcel_nb = models.Parcel.objects.count()
        ope.parcels.add(parcel)
        ope.parcels.clear()  # no signal raised... should resave
        models.Parcel.objects.filter(pk=parcel.pk).all()[0].save()
        # our parcel has no operation attached and should be deleted
        self.assertEqual(parcel_nb - 1, models.Parcel.objects.count())

    def testIndex(self):
        ope = create_operation(self.user, values={'year': 2042})
        self.assertEqual(ope.operation_code, 1)
        ope2 = create_operation(self.user, values={'year': 2042})
        self.assertEqual(ope2.operation_code, 2)
        ope = create_operation(self.user, values={'year': 0})
        self.assertEqual(ope.operation_code, 1)
        ope2 = create_operation(self.user, values={'year': 0})
        self.assertEqual(ope2.operation_code, 2)

    def test_cache_update(self):
        self.create_towns()
        operation = self.operations[0]
        town, ope_id = operation.cached_label.split(' | ')
        self.assertIn(town, (u'Intercommunal', u"Multi-town"))
        self.assertEqual(ope_id, 'OP2010-1')
        operation = models.Operation.objects.get(pk=operation.pk)
        operation.year = 2011
        operation.save()
        operation.towns.add(self.towns[0])
        operation = models.Operation.objects.get(pk=operation.pk)
        town, ope_id = operation.cached_label.split(' | ')
        self.assertEqual(ope_id, 'OP2011-1')
        self.assertEqual(town, self.towns[0].name)

    def test_cache_bulk_update(self):
        if settings.USE_SPATIALITE_FOR_TESTS:
            # using views - can only be tested with postgresql
            return

        operation = self.operations[0]
        init_parcel = self.create_parcel()[0]
        operation.parcels.add(init_parcel)

        from archaeological_context_records.models import ContextRecord
        cr_data = {'label': "Context record", "operation": operation,
                   'parcel': init_parcel,
                   'history_modifier': self.get_default_user()}
        cr = ContextRecord.objects.create(**cr_data)

        from archaeological_finds.models import BaseFind, Find, MaterialType
        bf_data = {
            'label': "Base find", 'history_modifier': self.get_default_user(),
            'context_record': cr
        }
        base_find = BaseFind.objects.create(**bf_data)
        find = Find.objects.create(
            history_modifier=self.get_default_user(),
            label='Find me'
        )
        find.base_finds.add(base_find)
        mat = MaterialType.objects.create(
            label='Adamentium', txt_idx='admentium', code='ADA')
        find.material_types.add(mat)

        class TestObj(object):
            def __init__(self):
                self.context_record_reached = []

            def reached(self, sender, **kwargs):
                instance = kwargs.get('instance')
                if sender == ContextRecord:
                    self.context_record_reached.append(instance)

        test_obj = TestObj()
        operation = models.Operation.objects.get(pk=operation.pk)
        operation.test_obj = test_obj
        operation.year = 2011
        operation.save()
        # bulk update of context records cached label gen don't have to be
        # reached
        self.assertEqual(len(test_obj.context_record_reached), 0)

        # verify the relevance of the update
        cr = ContextRecord.objects.get(pk=cr.pk)
        ope_id, parcel_sec, parcel_nb, cr_label = cr.cached_label.split(' | ')
        self.assertEqual(ope_id, '{}2011-1'.format(
            settings.ISHTAR_DEF_OPE_PREFIX))

        base_find = BaseFind.objects.get(pk=base_find.pk)
        op_code, idx = base_find.cache_short_id.split(' | ')
        self.assertEqual(op_code, 'OP2011-1')
        self.assertEqual(idx, '00001')
        op_code, mat_code, lbl, idx = base_find.cache_complete_id.split(' | ')
        self.assertEqual(op_code, 'OP2011-1')
        self.assertEqual(mat_code, 'ADA')
        self.assertEqual(lbl, 'Context record')
        self.assertEqual(idx, '00001')

        find = Find.objects.get(pk=find.pk)
        op_code_idx, lbl = find.cached_label.split(' | ')
        self.assertEqual(op_code_idx, 'OP2011-1-00001')
        self.assertEqual(lbl, 'Find me')

        operation = models.Operation.objects.get(pk=operation.pk)
        operation.code_patriarche = '666'
        operation.save()
        cr = ContextRecord.objects.get(pk=cr.pk)
        ope_id, parcel_sec, parcel_nb, cr_label = cr.cached_label.split(' | ')
        self.assertEqual(ope_id, '{}666'.format(settings.ISHTAR_OPE_PREFIX))

        base_find = BaseFind.objects.get(pk=base_find.pk)
        op_code, idx = base_find.cache_short_id.split(' | ')
        self.assertEqual(op_code, 'OA666')
        op_code, mat_code, lbl, idx = base_find.cache_complete_id.split(' | ')
        self.assertEqual(op_code, 'OA666')

        find = Find.objects.get(pk=find.pk)
        op_code_idx, lbl = find.cached_label.split(' | ')
        self.assertEqual(op_code_idx, 'OA666-00001')


class OperationSearchTest(TestCase, OperationInitTest):
    fixtures = [settings.ROOT_PATH +
                '../fixtures/initial_data-auth-fr.json',
                settings.ROOT_PATH +
                '../ishtar_common/fixtures/initial_data-fr.json',
                settings.ROOT_PATH +
                '../archaeological_files/fixtures/initial_data.json',
                settings.ROOT_PATH +
                '../archaeological_operations/fixtures/initial_data-fr.json']

    def setUp(self):
        IshtarSiteProfile.objects.get_or_create(
            slug='default', active=True)
        self.username, self.password, self.user = create_superuser()
        self.alt_username, self.alt_password, self.alt_user = create_user()
        self.alt_user.user_permissions.add(Permission.objects.get(
            codename='view_own_operation'))
        self.orgas = self.create_orgas(self.user)
        self.operations = self.create_operation(self.user, self.orgas[0])
        self.operations += self.create_operation(self.alt_user, self.orgas[0])
        self.item = self.operations[0]

    def test_base_search(self):
        c = Client()
        response = c.get(reverse('get-operation'), {'year': '2010'})
        # no result when no authentication
        self.assertTrue(not json.loads(response.content))
        c.login(username=self.username, password=self.password)
        response = c.get(reverse('get-operation'), {'year': '2010'})
        self.assertEqual(json.loads(response.content)['total'], 2)
        response = c.get(reverse('get-operation'),
                         {'operator': self.orgas[0].pk})
        result = json.loads(response.content)
        self.assertEqual(result['total'], 2)

    def create_relations(self):
        rel1 = models.RelationType.objects.create(
            symmetrical=True, label='Include', txt_idx='include')
        rel2 = models.RelationType.objects.create(
            symmetrical=False, label='Included', txt_idx='included',
            inverse_relation=rel1)
        models.RecordRelations.objects.create(
            left_record=self.operations[0],
            right_record=self.operations[1],
            relation_type=rel1)
        return rel1, rel2

    def testRelatedSearch(self):
        c = Client()
        rel1, rel2 = self.create_relations()
        self.operations[1].year = 2011
        self.operations[1].save()
        search = {'year': '2010', 'relation_types_0': rel2.pk}
        response = c.get(reverse('get-operation'), search)
        # no result when no authentification
        self.assertTrue(not json.loads(response.content))
        c.login(username=self.username, password=self.password)
        response = c.get(reverse('get-operation'), search)
        self.assertTrue(json.loads(response.content)['total'] == 2)

    def testHierarchicSearch(self):
        ope = self.operations[1]
        c = Client()

        neo = models.Period.objects.get(txt_idx='neolithic')
        final_neo = models.Period.objects.get(txt_idx='final_neolithic')
        recent_neo = models.Period.objects.get(txt_idx='recent_neolithic')
        ope.periods.add(final_neo)

        search = {'periods': final_neo.pk}

        # no result when no authentication
        response = c.get(reverse('get-operation'), search)
        self.assertEqual(response.status_code, 200)
        self.assertTrue(not json.loads(response.content))
        c.login(username=self.username, password=self.password)

        # one result for exact search
        response = c.get(reverse('get-operation'), search)
        self.assertEqual(response.status_code, 200)
        res = json.loads(response.content)
        self.assertTrue(res['total'] == 1)

        # no result for the brother
        search = {'periods': recent_neo.pk}
        response = c.get(reverse('get-operation'), search)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(json.loads(response.content)['total'], 0)

        # one result for the father
        search = {'periods': neo.pk}
        response = c.get(reverse('get-operation'), search)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(json.loads(response.content)['total'], 1)

    def testOwnSearch(self):
        c = Client()
        response = c.get(reverse('get-operation'), {'year': '2010'})
        # no result when no authentification
        self.assertTrue(not json.loads(response.content))
        c.login(username=self.alt_username, password=self.alt_password)
        response = c.get(reverse('get-operation'), {'year': '2010'})
        # only one "own" operation available
        self.assertTrue(json.loads(response.content)['total'] == 1)
        response = c.get(reverse('get-operation'),
                         {'operator': self.orgas[0].pk})
        self.assertTrue(json.loads(response.content)['total'] == 1)


def create_administrativact(user, operation):
    act_type, created = models.ActType.objects.get_or_create(
        txt_idx='act_type')
    dct = {'history_modifier': user,
           'act_type': act_type,
           'operation': operation,
           'signature_date': datetime.date(2014, 05, 12),
           'index': 322}
    adminact, created = models.AdministrativeAct.objects.get_or_create(**dct)
    return [act_type], [adminact]


class RegisterTest(TestCase, OperationInitTest):
    fixtures = [settings.ROOT_PATH +
                '../ishtar_common/fixtures/initial_data.json',
                settings.ROOT_PATH +
                '../archaeological_files/fixtures/initial_data.json',
                settings.ROOT_PATH +
                '../archaeological_operations/fixtures/initial_data-fr.json']

    def setUp(self):
        self.username, self.password, self.user = create_superuser()
        self.operations = self.create_operation(self.user)
        self.act_types, self.operations = create_administrativact(
            self.user, self.operations[0])

    def testSearch(self):
        c = Client()
        response = c.get(reverse('get-administrativeact'), {'year': '2014'})
        # no result when no authentication
        self.assertTrue(not json.loads(response.content))
        c.login(username=self.username, password=self.password)
        response = c.get(reverse('get-administrativeact'), {'year': '2014'})
        self.assertTrue(json.loads(response.content)['total'] == 1)
        response = c.get(reverse('get-administrativeact'), {'indexed': '2'})
        self.assertTrue(json.loads(response.content)['total'] == 1)


class OperationWizardCreationTest(WizardTest, OperationInitTest, TestCase):
    fixtures = [settings.ROOT_PATH +
                '../fixtures/initial_data-auth-fr.json',
                settings.ROOT_PATH +
                '../ishtar_common/fixtures/initial_data-fr.json',
                settings.ROOT_PATH +
                '../archaeological_files/fixtures/initial_data.json',
                settings.ROOT_PATH +
                '../archaeological_operations/fixtures/initial_data-fr.json']
    url_name = 'operation_creation'
    wizard_name = 'operation_wizard'
    steps = views.wizard_steps
    form_datas = [
        FormData(
            "Create a preventive diag",
            form_datas={
                'filechoice-operation_creation': {},
                'general-operation_creation': {
                    'operation_type': 1,  # preventive diag
                    'year': 2016},
                'townsgeneral-operation_creation': [],
                'parcelsgeneral-operation_creation': [],
            },
            ignored=('towns-operation_creation',
                     'parcels-operation_creation',
                     'preventive-operation_creation',)
        ),
        FormData(
            "Create another preventive diag with same parcel name",
            form_datas={
                'filechoice-operation_creation': {},
                'general-operation_creation': {
                    'operation_type': 1,  # preventive diag
                    'year': 2016},
                'townsgeneral-operation_creation': [],
                'parcelsgeneral-operation_creation': [],
            },
            ignored=('towns-operation_creation',
                     'parcels-operation_creation',
                     'preventive-operation_creation',)
        )
    ]

    def pre_wizard(self):
        profile, created = IshtarSiteProfile.objects.get_or_create(
            slug='default', active=True)
        profile.files = True
        profile.save()

        if 'townsgeneral-operation_creation' not in \
                self.form_datas[0].form_datas:
            return super(OperationWizardCreationTest, self).pre_wizard()
        town = self.create_towns()[0]
        town_data = {'town': town.pk}
        self.form_datas[0].form_datas[
            'townsgeneral-operation_creation'].append(town_data)
        self.form_datas[1].form_datas[
            'townsgeneral-operation_creation'].append(town_data)
        parcel_data = {
            'town': town.pk, 'year': 2017, 'section': 'S',
            'parcel_number': '42'}
        self.form_datas[0].form_datas[
            'parcelsgeneral-operation_creation'].append(parcel_data)
        self.form_datas[1].form_datas[
            'parcelsgeneral-operation_creation'].append(parcel_data)
        self.operation_number = models.Operation.objects.count()
        self.parcel_number = models.Parcel.objects.count()
        super(OperationWizardCreationTest, self).pre_wizard()

    def post_wizard(self):
        self.assertEqual(models.Operation.objects.count(),
                         self.operation_number + 2)
        self.assertEqual(models.Parcel.objects.count(),
                         self.parcel_number + 2)


class OperationWizardModifTest(WizardTest, OperationInitTest, TestCase):
    fixtures = OperationWizardCreationTest.fixtures
    url_name = 'operation_modification'
    wizard_name = url_name + '_wizard'
    steps = views.operation_modif_wizard_steps
    base_ignored_steps = (
        'archaeologicalsite-operation_modification',
        'preventive-operation_modification',
        'preventivediag-operation_modification',
        'towns-operation_modification',
        'parcels-operation_modification',
        'remains-operation_modification',
        'periods-operation_modification',
        'relations-operation_modification',
        'abstract-operation_modification',)
    form_datas = [
        FormData(
            "Update an operation",
            form_datas={
                'selec-operation_modification': {},
                'general-operation_modification': {
                    'operation_type': 2,
                    'year': 2017},
                'townsgeneral-operation_modification': [],
                'parcelsgeneral-operation_modification': [],
            },
            ignored=base_ignored_steps
        ),
        FormData(
            "Operation: try to remove a parcel with attached context record",
            form_datas={
                'selec-operation_modification': {},
                'general-operation_modification': {
                    'operation_type': 2,
                    'year': 2017},
                'townsgeneral-operation_modification': [],
                'parcelsgeneral-operation_modification': [],
            },
            ignored=base_ignored_steps
        ),
        FormData(
            "Operation: remove a parcel with no attached context record",
            form_datas={
                'selec-operation_modification': {},
                'general-operation_modification': {
                    'operation_type': 2,
                    'year': 2017},
                'townsgeneral-operation_modification': [],
                'parcelsgeneral-operation_modification': [],
            },
            ignored=base_ignored_steps
        ),
    ]

    def pre_wizard(self):
        self.create_operation()
        operation = self.operations[0]
        init_town = self.create_towns()[0]
        operation.towns.add(init_town)
        init_parcel = self.create_parcel()[0]
        operation.parcels.add(init_parcel)

        from archaeological_context_records.models import ContextRecord
        cr_data = {'label': "Context record", "operation": operation,
                   'parcel': init_parcel,
                   'history_modifier': self.get_default_user()}
        self.cr = ContextRecord.objects.create(**cr_data)

        data = self.form_datas[0].form_datas
        data2 = self.form_datas[1].form_datas
        data3 = self.form_datas[2].form_datas

        data['selec-operation_modification']['pk'] = operation.pk
        data2['selec-operation_modification']['pk'] = operation.pk
        data3['selec-operation_modification']['pk'] = operation.pk

        town = self.create_towns(
            datas={'numero_insee': '67890', 'name': 'Twin Peaks'})[-1]
        towns = [{'town': town.pk}, {'town': init_town.pk}]
        data['townsgeneral-operation_modification'] = towns
        data2['townsgeneral-operation_modification'] = towns
        data3['townsgeneral-operation_modification'] = towns

        parcel_data = {
            'town': town.pk, 'year': 2017, 'section': 'S',
            'parcel_number': '42'}
        data['parcelsgeneral-operation_modification'].append(parcel_data)
        data2['parcelsgeneral-operation_modification'].append(parcel_data)
        data3['parcelsgeneral-operation_modification'].append(parcel_data)

        parcel_data_2 = {
            'town': init_parcel.town.pk, 'year': init_parcel.year or '',
            'section': init_parcel.section,
            'parcel_number': init_parcel.parcel_number}
        data['parcelsgeneral-operation_modification'].append(parcel_data_2)
        # no init parcel for data2 and data3

        self.operation_number = models.Operation.objects.count()
        self.parcel_number = models.Parcel.objects.count()

        def post_first_wizard(test_object, final_step_response):
            test_object.assertEqual(models.Operation.objects.count(),
                                    test_object.operation_number)
            operation = models.Operation.objects.get(
                pk=test_object.operations[0].pk)
            test_object.assertEqual(operation.operation_type.pk, 2)
            test_object.assertEqual(operation.year, 2017)
            test_object.assertEqual(models.Parcel.objects.count(),
                                    test_object.parcel_number + 1)
            test_object.assertEqual(operation.parcels.count(),
                                    test_object.parcel_number + 1)

        def post_second_wizard(test_object, final_step_response):
            test_object.assertEqual(models.Operation.objects.count(),
                                    test_object.operation_number)
            operation = models.Operation.objects.get(
                pk=test_object.operations[0].pk)
            test_object.assertEqual(operation.operation_type.pk, 2)
            test_object.assertEqual(operation.year, 2017)
            test_object.assertEqual(models.Parcel.objects.count(),
                                    test_object.parcel_number + 1)
            # the init parcel is not submited but have a context record
            # the init parcel is not detached from the operation
            test_object.assertEqual(operation.parcels.count(),
                                    test_object.parcel_number + 1)

        def pre_third_wizard(test_object):
            parcel_nb = models.Parcel.objects.count()
            test_object.cr.delete()
            test_object.assertEqual(
                parcel_nb, models.Parcel.objects.count())

        def post_third_wizard(test_object, final_step_response):
            test_object.assertEqual(models.Operation.objects.count(),
                                    test_object.operation_number)
            operation = models.Operation.objects.get(
                pk=test_object.operations[0].pk)
            test_object.assertEqual(operation.operation_type.pk, 2)
            test_object.assertEqual(operation.year, 2017)
            # with no attach the parcel is deleted
            test_object.assertEqual(operation.parcels.count(),
                                    test_object.parcel_number)
            test_object.assertEqual(models.Parcel.objects.count(),
                                    test_object.parcel_number)

        self.form_datas[0].extra_tests = [post_first_wizard]
        self.form_datas[1].extra_tests = [post_second_wizard]
        self.form_datas[2].pre_tests = [pre_third_wizard]
        self.form_datas[2].extra_tests = [post_third_wizard]
        super(OperationWizardModifTest, self).pre_wizard()


class OperationWizardDeleteTest(OperationWizardCreationTest):
    fixtures = OperationWizardCreationTest.fixtures
    url_name = 'operation_deletion'
    wizard_name = 'operation_deletion_wizard'
    steps = views.operation_deletion_steps
    form_datas = [
        FormData(
            "Wizard deletion test",
            form_datas={
                'selec-operation_deletion': {'pk': None},
            }
        )
    ]

    def pass_test(self):
        if not settings.SOUTH_TESTS_MIGRATE:
            # with no migration the views are not created
            return True

    def pre_wizard(self):
        self.ope = self.get_default_operation(force=True)
        self.ope.parcels.add(self.create_parcel()[0])
        self.parcel_nb = models.Parcel.objects.count()
        self.form_datas[0].form_datas['selec-operation_deletion']['pk'] = \
            self.ope.pk
        self.operation_number = models.Operation.objects.count()
        super(OperationWizardDeleteTest, self).pre_wizard()

    def post_wizard(self):
        self.assertEqual(self.operation_number - 1,
                         models.Operation.objects.count())
        # associated parcel removed
        self.assertEqual(self.parcel_nb - 1, models.Parcel.objects.count())


class OperationWizardClosingTest(OperationWizardCreationTest):
    fixtures = OperationWizardCreationTest.fixtures
    url_name = 'operation_closing'
    wizard_name = 'operation_closing_wizard'
    steps = views.operation_closing_steps
    form_datas = [
        FormData(
            "Wizard closing test",
            form_datas={
                'selec-operation_closing': {'pk': None},
                'date-operation_closing': {'end_date': '2016-01-01'},
            }
        )
    ]

    def pre_wizard(self):
        self.ope = self.get_default_operation()
        self.form_datas[0].form_datas['selec-operation_closing']['pk'] = \
            self.ope.pk
        self.assertTrue(self.ope.is_active())
        super(OperationWizardClosingTest, self).pre_wizard()

    def post_wizard(self):
        ope = models.Operation.objects.get(pk=self.ope.pk)
        self.assertFalse(ope.is_active())
        self.assertEqual(
            ope.closing()['date'].strftime('%Y-%d-%m'),
            self.form_datas[0].form_datas['date-operation_closing']['end_date']
        )


class OperationAdminActWizardCreationTest(WizardTest, OperationInitTest,
                                          TestCase):
    fixtures = [settings.ROOT_PATH +
                '../fixtures/initial_data-auth-fr.json',
                settings.ROOT_PATH +
                '../ishtar_common/fixtures/initial_data-fr.json',
                settings.ROOT_PATH +
                '../archaeological_files/fixtures/initial_data.json',
                settings.ROOT_PATH +
                '../archaeological_operations/fixtures/initial_data-fr.json']
    url_name = 'operation_administrativeactop'
    wizard_name = 'operation_administrative_act_wizard'
    steps = views.administrativeactop_steps
    form_datas = [
        FormData(
            "Admin act creation",
            form_datas={
                'selec-operation_administrativeactop': {},
                'administrativeact-operation_administrativeactop': {
                    'signature_date': str(datetime.date.today())
                }
            },
        )
    ]

    def pre_wizard(self):
        ope = self.get_default_operation()
        self.number = models.AdministrativeAct.objects.count()

        data = self.form_datas[0].form_datas
        data['selec-operation_administrativeactop']['pk'] = ope.pk
        act = models.ActType.objects.filter(intented_to='O').all()[0].pk

        data['administrativeact-operation_administrativeactop'][
            'act_type'] = act
        super(OperationAdminActWizardCreationTest, self).pre_wizard()

    def post_wizard(self):
        self.assertEqual(models.AdministrativeAct.objects.count(),
                         self.number + 1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2012-2013 Étienne Loks  <etienne.loks_AT_peacefrogsDOTnet>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# See the file COPYING for details.

from ishtar_common.models import get_current_profile


class SectionItem:
    def __init__(self, idx, label, childs=[], profile_restriction=None, css=''):
        self.idx = idx
        self.label = label
        self.childs = childs
        self.available = False
        self.items = {}
        self.profile_restriction = profile_restriction
        self.css = css

    def check_profile_restriction(self):
        if self.profile_restriction:
            profile = get_current_profile()
            if not getattr(profile, self.profile_restriction):
                return False
        return True

    def can_be_available(self, user, session=None):
        if not self.check_profile_restriction():
            return False
        for child in self.childs:
            if child.can_be_available(user, session=session):
                return True
        return False

    def is_available(self, user, obj=None, session=None):
        for child in self.childs:
            if child.is_available(user, obj, session=session):
                return True
        return False

    def set_items(self, user, items, current_action=None, session=None):
        selected = None
        if user:
            self.available = self.can_be_available(user, session=session)
        for child in self.childs:
            selected = child.set_items(user, items, current_action,
                                       session=session) or selected
            items[child.idx] = child
        return selected


class MenuItem:
    def __init__(self, idx, label, model=None, access_controls=[],
                 profile_restriction=None, css=''):
        self.idx = idx
        self.label = label
        self.model = model
        self.access_controls = access_controls
        self.available = False
        self.profile_restriction = profile_restriction
        self.css = css
        if not self.check_profile_restriction():
            return False

    def check_profile_restriction(self):
        if self.profile_restriction:
            profile = get_current_profile()
            if not getattr(profile, self.profile_restriction):
                return False
        return True

    def can_be_available(self, user, session=None):
        if not self.check_profile_restriction():
            return False
        if not self.access_controls:
            return True
        if not hasattr(user, 'ishtaruser'):
            return False
        # manage by specific idx - person type
        if user.ishtaruser.has_right(self.idx, session=session):
            return True
        prefix = (self.model._meta.app_label + '.') if self.model else ''
        for access_control in self.access_controls:
            # check by person type
            if user.ishtaruser.has_right(access_control, session=session):
                return True
            access_control = prefix + access_control
            # check by specific access control
            if user.ishtaruser.has_perm(access_control, self.model,
                                        session=session) or \
               access_control in user.get_group_permissions():
                return True
        return False

    def is_available(self, user, obj=None, session=None):
        if not self.check_profile_restriction():
            return False
        if not self.access_controls:
            return True
        prefix = (self.model._meta.app_label + '.') if self.model else ''
        for access_control in self.access_controls:
            access_control = prefix + access_control
            if user.has_perm(access_control, self.model, obj=obj):
                #              session=session):
                return True
        # manage by person type
        if hasattr(user, 'ishtaruser'):
            if user.ishtaruser.has_right(self.idx, session=session):
                return True
        return False

    def set_items(self, user, items, current_action=None, session=None):
        if user:
            self.available = self.can_be_available(user, session=session)
        if self.idx == current_action:
            return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2012-2016 Étienne Loks  <etienne.loks_AT_peacefrogsDOTnet>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# See the file COPYING for details.

from django.contrib import admin

from ishtar_common.admin import HistorizedObjectAdmin, GeneralTypeAdmin

import models


class BaseFindAdmin(HistorizedObjectAdmin):
    list_display = ('label', 'context_record', 'batch')
    search_fields = ('label', 'context_record__parcel__operation__name',)
    model = models.BaseFind

admin.site.register(models.BaseFind, BaseFindAdmin)


class FindAdmin(HistorizedObjectAdmin):
    list_display = ('label', 'dating', 'volume', 'weight',
                    'find_number',)
    search_fields = ('label', "datings__period__label")
    model = models.Find

admin.site.register(models.Find, FindAdmin)


class FindSourceAdmin(admin.ModelAdmin):
    list_display = ('find', 'title', 'source_type',)
    list_filter = ('source_type',)
    search_fields = ('title', )
    model = models.FindSource

admin.site.register(models.FindSource, FindSourceAdmin)


class PropertyAdmin(admin.ModelAdmin):
    list_display = ['find', 'person', 'start_date', 'end_date']
    search_fields = ('find__label', 'person__name')
    model = models.Property

admin.site.register(models.Property, PropertyAdmin)


class TreatmentAdmin(HistorizedObjectAdmin):
    list_display = ('location', 'treatment_types_lbl', 'container', 'person')
    model = models.Treatment

admin.site.register(models.Treatment, TreatmentAdmin)


class TreatmentFileAdmin(HistorizedObjectAdmin):
    list_display = ('year', 'index', 'name', 'internal_reference')
    search_fields = ('cached_label',)
    model = models.TreatmentFile

admin.site.register(models.TreatmentFile, TreatmentFileAdmin)


class TreatmentSourceAdmin(admin.ModelAdmin):
    list_display = ('treatment', 'title', 'source_type',)
    list_filter = ('source_type',)
    search_fields = ('title',)
    model = models.TreatmentSource

admin.site.register(models.TreatmentSource, TreatmentSourceAdmin)


class HierarchicalTypeAdmin(GeneralTypeAdmin):
    list_display = ['label', 'txt_idx', 'parent', 'available', 'comment']

admin.site.register(models.ObjectType, HierarchicalTypeAdmin)


class MaterialTypeAdmin(HierarchicalTypeAdmin):
    list_display = HierarchicalTypeAdmin.list_display + ['recommendation']
    search_fields = ('label', 'parent__label', 'comment',)

admin.site.register(models.MaterialType, MaterialTypeAdmin)


class TreatmentTypeAdmin(admin.ModelAdmin):
    list_display = HierarchicalTypeAdmin.list_display + [
        'order', 'virtual', 'upstream_is_many', 'downstream_is_many']
    model = models.TreatmentType
admin.site.register(models.TreatmentType, TreatmentTypeAdmin)

general_models = [
    models.ConservatoryState, models.RemarkabilityType,
    models.PreservationType, models.IntegrityType,
    models.TreatmentFileType, models.TreatmentState,
    models.BatchType
]
for model in general_models:
    admin.site.register(model, GeneralTypeAdmin)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2012-2014 Étienne Loks  <etienne.loks_AT_peacefrogsDOTnet>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# See the file COPYING for details.

from django.conf import settings
from django.contrib import admin

from ishtar_common.admin import HistorizedObjectAdmin, GeneralTypeAdmin

import models


class AdministrativeActAdmin(HistorizedObjectAdmin):
    list_display = ('year', 'index', 'operation', 'associated_file',
                    'act_type')
    list_filter = ('act_type',)
    search_fields = ('year', 'index')
    readonly_fields = ('in_charge', 'operator', 'scientist', 'signatory',
                       'associated_file', 'imports',
                       'departments_label', 'towns_label',
                       'history_modifier', 'history_creator')
    model = models.AdministrativeAct

admin.site.register(models.AdministrativeAct, AdministrativeActAdmin)


class PeriodAdmin(admin.ModelAdmin):
    list_display = ('label', 'start_date', 'end_date', 'parent', 'available',
                    'order')
    list_filter = ('parent',)
    model = models.Period

admin.site.register(models.Period, PeriodAdmin)


class ArchaeologicalSiteAdmin(HistorizedObjectAdmin):
    list_display = ('name', 'reference')
    search_fields = ('name', 'reference')
    model = models.ArchaeologicalSite

admin.site.register(models.ArchaeologicalSite, ArchaeologicalSiteAdmin)


class OperationAdmin(HistorizedObjectAdmin):
    list_display = ['year', 'operation_code', 'start_date',
                    'excavation_end_date', 'end_date',
                    'operation_type']
    list_filter = ("year", "operation_type",)
    search_fields = ['towns__name', 'operation_code']
    if settings.COUNTRY == 'fr':
        list_display += ['code_patriarche']
        search_fields += ['code_patriarche']
    model = models.Operation

admin.site.register(models.Operation, OperationAdmin)


class OperationSourceAdmin(admin.ModelAdmin):
    list_display = ('operation', 'title', 'source_type',)
    list_filter = ('source_type',)
    search_fields = ('title', 'operation__common_name')
    model = models.OperationSource

admin.site.register(models.OperationSource, OperationSourceAdmin)


class ParcelAdmin(HistorizedObjectAdmin):
    list_display = ['section', 'parcel_number', 'operation', 'associated_file']
    search_fields = ('operation__name',)
    model = models.Parcel

admin.site.register(models.Parcel, ParcelAdmin)


class RecordRelationsAdmin(admin.ModelAdmin):
    list_display = ('left_record', 'relation_type', 'right_record')
    list_filter = ('relation_type',)

admin.site.register(models.RecordRelations, RecordRelationsAdmin)


class RelationTypeAdmin(admin.ModelAdmin):
    list_display = ('label', 'txt_idx', 'tiny_label', 'available',
                    'symmetrical', 'inverse_relation', 'comment')

admin.site.register(models.RelationType, RelationTypeAdmin)


class ActTypeAdmin(GeneralTypeAdmin):
    list_filter = ('intented_to',)
    list_display = ['label', 'txt_idx', 'available', 'intented_to']

admin.site.register(models.ActType, ActTypeAdmin)


class ReportStateAdmin(GeneralTypeAdmin):
    list_display = ['label', 'txt_idx', 'available', 'order', 'comment']

admin.site.register(models.ReportState, ReportStateAdmin)


general_models = [models.RemainType]
for model in general_models:
    admin.site.register(model, GeneralTypeAdmin)

basic_models = [models.ParcelOwner]
for model in basic_models:
    admin.site.register(model)



from ishtar_common import models


def convert_field(field_name):
    if field_name.startswith('find__'):
        field_name = field_name[len('find__'):]
    else:
        field_name = "base_finds__" + field_name
    return field_name


def refac_types(types):
    find_model, created = models.ImporterModel.objects.get_or_create(
        klass='archaeological_finds.models_finds.Find',
        defaults={'name': 'Find'}
    )
    for tpe in types:
        for col in tpe.columns.all():
            for field in col.duplicate_fields.all():
                new_field_name = convert_field(field.field_name)
                field.field_name = new_field_name
                field.save()
            for field in col.targets.all():
                new_field_name = convert_field(field.target)
                field.target = new_field_name
                field.save()
        tpe.associated_models = find_model
        tpe.save()


types = list(models.ImporterType.objects.filter(
    associated_models=models.ImporterModel.objects.get(
        klass='archaeological_finds.models.BaseFind')).all())

refac_types(types)
